//*****************************************************************************
// das_ad_r.c module contains routines to process messages sent by a client to 
// perform the function of the 'das' library routines.  The true operation of 
// the functions may not be identical to the actual low level 'das' routine 
// because of the added feature of supporting multiple applications using the 
// same adapter.
//*****************************************************************************
// The following is the list of functions in this module.
//
//     Message processing function     DAS driver function 
//     ---------------------------     -------------------
//         das_ad_sync_msg          -     das_ad_sync
//         das_ad_enqueue_msg       -     das_ad_enqueue
//         das_ad_read_msg          -     das_ad_read
//                                  -     das_ad_cread
//         das_ad_set_scan_msg      -     das_ad_set_scan
//         das_ad_term_msg          -     das_ad_term
//         das_ad_sam_msg           -     das_ad_set_sample
//         das_ad_get_rng_msg       -     das_ad_get_range
//         das_ad_set_rng_msg       -     das_ad_set_range
//         das_ad_set_trig_msg      -     das_ad_set_trig
//         das_ad_get_trig_msg      -     das_ad_get_trig
//
//         excl_das_ad_enqueue_msg  -     used for exclusive access to manager.
//*****************************************************************************
#include <stdlib.h>
#include <errno.h>
#include <malloc.h>
#include <sys/types.h>       
#include <sys/kernel.h>
#include <sys/sendmx.h>

#include <das.h>
#include "struct.h"
#include "extern.h"
#include "proto.h"
#include "define.h"

//*****************************************************************************
// das_ad_sync_msg() -  Retrieve one sample and send it to the requesting 
//                      client.  
//*****************************************************************************
int das_ad_sync_msg(pid_t pid)
{  
   struct ocb *ocb;
   struct _mxfer_entry mx[2];
   das_t data;

                     // Get 'ocb' assoicated with calling process.
   if ((ocb = get_ocb(pid, io_msg.das_ad_sync.fd)) == -1)
   {
      io_msg.das_ad_sync_reply.status = EBADF;
      return(REPLY);
   }
                     // Check access mode.
   if ((ocb->mode & S_IREAD) == 0)
   {
      io_msg.status = EACCES;
      return(REPLY);
   }
                     // Can't queue this request.  
   if (ocb->req_t == A_D)
   {
      io_msg.status = EBUSY;
      return(REPLY);
   }
   io_msg.status = EOK;

                     // Get the conversion.  
   if (in_use == SINGLE_USER)
   {
      if (das_ad_sync(das_fd, io_msg.das_ad_sync.chnl, (char *)&data) == -1)
         io_msg.status = errno;
   }
   else   
      data = das_res_buf[io_msg.das_ad_sync.chnl]; 
   io_msg.das_ad_sync_reply.n_bytes = d_info.ad_res;
   
                     // Setup reply header.
   _setmx(&mx[0], &io_msg, sizeof(io_msg.das_ad_sync_reply) - 
            sizeof(io_msg.das_ad_sync_reply.data));
                     // Setup reply data.
   _setmx(&mx[1], &data, d_info.ad_res);
   
   Replymx(pid, 2, &mx);
   return(NO_REPLY);
}

//*****************************************************************************
// das_ad_enqueue_msg() -  Enqueue a A/D request from a client.
//*****************************************************************************
int das_ad_enqueue_msg(pid_t pid)
{
   struct ocb *ocb;
   struct das_res_queue *tmp;

                     // Get 'ocb' assoicated with calling process.      
   if ((ocb = get_ocb(pid, io_msg.das_ad_enque.fd)) == -1)
   {
      io_msg.status = EBADF;
      return(REPLY);
   }
                     // Check access mode.
   if ((ocb->mode & S_IREAD) == 0)
   {
      io_msg.status = EACCES;
      return(REPLY);
   }
                     // Make sure a D/A conversion is not is progress.   
   if (ocb->req_t == D_A)
   {
      io_msg.status = EBUSY;
      return(REPLY);
   }
                     // Recycle mode not supported in acquistion manager.
   if (io_msg.das_ad_enque.mode & DAS_RECYCLE)
   {
      io_msg.status = EINVAL;
      return(REPLY);
   }
                     // Check request size with buffer size.
   if (io_msg.das_ad_enque.n_bytes > gbuf_size)
   {
      io_msg.status = ENOMEM;
      return(REPLY);
   }
                     // Are we opened for O_EXCL
   if (in_use == SINGLE_USER)
   {
      ocb->pid = pid;
      return(excl_das_ad_enqueue_msg(ocb));
   }
                     // Create que entry
   if ((tmp = das_res_create_entry()) == NULL)
   {
      io_msg.status = errno;
      return(REPLY);
   }
                     // Get data buffer.
   if ((tmp->mem_que = das_get_buf(ocb)) == NULL)
   {
      io_msg.status = ENOMEM;
      return(REPLY);
   }
                     // Starting pointer of data buffer.
   tmp->cp = (das_t *)tmp->mem_que->buffer->bufp;              
                                             // Number of conversions to do.
   tmp->nconv = io_msg.das_ad_enque.n_bytes / d_info.ad_res;
   tmp->cur_conv = tmp->nconv;               // Count down for each conversion.
   tmp->fin_proxy = io_msg.das_ad_enque.proxy;// Proxy to notify client.
   tmp->chnl = ocb->ad_seq.seq[0];           // Get first chnl in scan queue.
   ocb->req_t = A_D;                         // Set request type.
                     // Set queue request type to POLL operation so that 
                     // we kown to unblock client later.
   if (io_msg.das_ad_enque.mode == DAS_DMA_SYNC)
      tmp->que_req_t = POLL;
   ocb->ad.n_rent++;
                     // Place entry in queue.
                     // If request is first in queue, then start the conversions
   if (das_res_link(tmp, &ocb->ad.r_hd, &ocb->ad.r_tl))
   {
      ocb->chnl_ent = 0;                     // Scan que entry #.
      ocb->t_id = das_add_tick_list(ocb, ocb->ad.ticks); // Add to time queue.
   }
   io_msg.status = EOK;
                     // Block client if doing POLL operation.
   if (tmp->que_req_t == POLL)               
      return(NO_REPLY);
   return(REPLY);
}

//*****************************************************************************
// das_ad_read_msg() -  Retrieve the stored data from the A/D conversions 
//                      previously requested by client.
//*****************************************************************************
int das_ad_read_msg(pid_t pid)
{
   struct ocb *ocb;
   struct _mxfer_entry mx[2];

                     // Get 'ocb' assoicated with calling process.
   if ((ocb = get_ocb(pid, io_msg.das_ad_read.fd)) == -1)
   {
      io_msg.status = EBADF;
      return(REPLY);
   }
                     // Are there any requests finished or pending?  
   if (ocb->ad.n_rent == 0 && ocb->ad.n_fent == 0)
   {
      io_msg.status = EDEADLK;
      return(REPLY);
   }
   io_msg.status = EOK;
   
                     // Is this a das_ad_cread message (non-blocking).
   if (io_msg.das_msg.request == _DAS_AD_CREAD)
   {
      if (ocb->ad.n_fent == 0)
      {
         io_msg.das_ad_read_reply.n_bytes = 0;
         return(REPLY);
      }
      Creceive(ocb->ad.f_hd->fin_proxy, 0, 0); 
   }                 // Has request finished yet.
   else if (ocb->ad.n_fent == 0)
   {
      ocb->ad.reuse = io_msg.das_ad_read.mode;
      ocb->ad.r_hd->que_req_t |= READ_BLK;   // Mark this request a read blocked
      return(NO_REPLY);
   }

   if (in_use == SINGLE_USER)                // dequeue request from direct lib.
      das_ad_dequeue(das_fd, 0);
                     // Can't reuse request if client has started a D/A request.
   if (io_msg.das_ad_read.mode == DAS_REUSE && ocb->req_t == D_A)
      io_msg.status = EBUSY;
                     // Setup reply header.
   io_msg.das_ad_read_reply.n_bytes = ocb->ad.f_hd->nconv * d_info.ad_res;
   _setmx(&mx[0], &io_msg, sizeof(io_msg.das_ad_read_reply) - 
            sizeof(io_msg.das_ad_read_reply.data));
                     // Setup data buffer to send.
   _setmx(&mx[1], ocb->ad.f_hd->mem_que->buffer->bufp, d_info.ad_res * 
            ocb->ad.f_hd->nconv);
   
   Replymx(pid, 2, &mx);                     // Reply with header and data.
   ocb->ad.n_fent--;
                     // Re-enqueue request 
   if (io_msg.das_ad_read.mode == DAS_REUSE && ocb->req_t != D_A)
   {
      if (in_use == SINGLE_USER)
      {
         ocb->ad.n_rent++;
         das_ad_enqueue(das_fd, 0, ocb->ad.f_hd->mem_que->buffer, 
                        io_msg.das_ad_read_reply.n_bytes, ocb->ad.f_hd->mode,
                        hw_adproxy);
                     // Place entry in que.
         das_res_link(ocb->ad.f_hd, &ocb->ad.r_hd, &ocb->ad.r_tl);
         das_res_unlink(KEEP, &ocb->ad.f_hd, &ocb->ad.f_tl);
         ocb->ad.r_tl->nptr = NULL;
         ocb->ad.n_rent++;
         return(NO_REPLY);
      }              // Starting pointer of data buffer. 
      ocb->ad.f_hd->cp = (das_t *)ocb->ad.f_hd->mem_que->buffer->bufp;         
                     // Number of conversions to do.
      ocb->ad.f_hd->cur_conv = ocb->ad.f_hd->nconv;
      ocb->ad.f_hd->chnl = ocb->ad_seq.seq[0]; // Get first chnl in scan queue.
      ocb->req_t = A_D;                        // Set request type.
                     // Place entry in queue.
                     // If request is first in queue, then start the conversion
      ocb->ad.n_rent++;
      if (das_res_link(ocb->ad.f_hd, &ocb->ad.r_hd, &ocb->ad.r_tl))
      {
         ocb->chnl_ent = 0;                     // Scan que entry #.
         ocb->t_id = das_add_tick_list(ocb, ocb->ad.ticks); // Add to time queue.
      }
      das_res_unlink(KEEP, &ocb->ad.f_hd, &ocb->ad.f_tl);
      ocb->ad.r_tl->nptr = NULL;
      return(NO_REPLY);
   }
   das_rel_buf(ocb->ad.f_hd->mem_que, ocb);  // Release data buffer.
                     // Remove from queue.
   das_res_unlink(FREE, &ocb->ad.f_hd, &ocb->ad.f_tl);
   return(NO_REPLY);
}

//*****************************************************************************
// das_ad_set_scan_msg() - Set block scan entries for A/D conversions.
//*****************************************************************************
int das_ad_set_scan_msg(pid_t pid)
{
   struct ocb *ocb;
   int seq;

                     // Get 'ocb' assoicated with calling process.   
   if ((ocb = get_ocb(pid, io_msg.das_ad_scan.fd)) == -1)
   {
      io_msg.status = EBADF;
      return(REPLY);
   }
                     // Check access mode & access flags.
   if ((ocb->mode & S_IREAD) == 0)
   {
      io_msg.status = EACCES;
      return(REPLY);
   }
   io_msg.status = EOK;
                     // Get the scan list from client.
   if (Readmsg(pid, sizeof(io_msg.das_ad_scan) - 
               sizeof(io_msg.das_ad_scan.data) -
			   sizeof(io_msg.das_ad_scan.n_entry), 
               &ocb->ad_seq, sizeof(struct das_seq)) == -1)
      io_msg.status = errno;
   else
   {                 // Opened in O_EXCL mode - call lib function.
      if (in_use == SINGLE_USER)
      {
         io_msg.status = EOK;
         if (das_ad_set_scan(das_fd, &ocb->ad_seq.seq, 
                              ocb->ad_seq.n_entries) == -1)
            io_msg.status = errno;
         return(REPLY);
      }
      else
      {              // Check for valid scan queue.
         if (ocb->ad_seq.n_entries > NUM_ENT)
            io_msg.status = EINVAL;
         else
         {
            for (seq = 0 ; seq < ocb->ad_seq.n_entries; seq++)
               if (ocb->ad_seq.seq[seq] >= num_ai)
               {
                  io_msg.status = EINVAL;
                  break;
               }
         }
      }
   }
   return(REPLY);
}

//*****************************************************************************
// das_ad_term_msg() -  Terminate all current, pending and finished A/D
//                      requests.
//*****************************************************************************
int das_ad_term_msg(pid_t pid)
{
   struct ocb *ocb;

                     // Get 'ocb' assoicated with calling process.
   if ((ocb = get_ocb(pid, io_msg.das_ad_term.fd)) == -1)
   {
      io_msg.status = EBADF;
      return(REPLY);
   }                                  
                     
   if (in_use == SINGLE_USER)                // Opened for O_EXCL
      das_ad_term(das_fd, 0);
   else
   {
      if (ocb->req_t == A_D)                 // A/D operation in progress?
         das_rm_tick_list(ocb->t_id);        // Stop current request.
   }
                     // Clear pending requests.
   while (ocb->ad.r_hd)
   {
      das_rel_buf(ocb->ad.r_hd->mem_que, ocb);        
      das_res_unlink(FREE, &ocb->ad.r_hd, &ocb->ad.r_tl);
   }
                     // Clear finished requests.
   while (ocb->ad.f_hd)
   {
      das_rel_buf(ocb->ad.f_hd->mem_que, ocb);        
      das_res_unlink(FREE, &ocb->ad.f_hd, &ocb->ad.f_tl);
   }
   ocb->req_t = NONE;
   ocb->ad.n_rent = 0;
   ocb->ad.n_fent = 0;
   io_msg.status = EOK;
   return(REPLY);
}

//*****************************************************************************
// das_ad_sam_msg() -   Set the sampling rate for A/D conversions.
//*****************************************************************************
int das_ad_sam_msg(pid_t pid)
{
   struct ocb *ocb;

                     // Get 'ocb' assoicated with calling process.   
   if ((ocb = get_ocb(pid, io_msg.das_ad_sam.fd)) == -1)
   {
      io_msg.status = EBADF;
      return(REPLY);
   }
                     // Check access mode & access flags.
   if ((ocb->mode & S_IREAD) == 0)
   {
      io_msg.status = EACCES;
      return(REPLY);
   }
   io_msg.status = EOK;   
   if (in_use == SINGLE_USER)                // Opened for O_EXCL.
   {
      if (das_ad_set_sample(das_fd, io_msg.das_ad_sam.chnl,                   
                              io_msg.das_ad_sam.ticks) == -1)
         io_msg.status = errno;
   }     
   else
   {                 // Check valid sampling rate.
      if ((io_msg.das_ad_sam.ticks < MIN_SAMPLE_TICK) ||
            (io_msg.das_ad_sam.ticks > MAX_SAMPLE_TICK))
      {
         io_msg.status = EINVAL;
         return(REPLY);
      }
      ocb->ad.ticks = io_msg.das_ad_sam.ticks;
      if (ocb->req_t == A_D)                 // If a request is in progress.
      {
         das_rm_tick_list(ocb->t_id);                       // Remove old entry
         ocb->t_id = das_add_tick_list(ocb, ocb->ad.ticks); // Add new entry 
      }  
   }
   return(REPLY);
}

//*****************************************************************************
// das_ad_set_rng_msg() -  Set the current range for the specified channel 
//                         (Must be root or have a euid of 0, or O_EXCL).
//*****************************************************************************
int das_ad_set_rng_msg(pid_t pid)
{
   struct ocb *ocb;
   uid_t uid;
   gid_t gid;

                     // Get 'ocb' assoicated with calling process.
   if ((ocb = get_ocb(pid, io_msg.das_ad_set_range.fd)) == -1)
   {
      io_msg.status = EBADF;
      return(REPLY);
   }
   io_msg.status = EOK;

   if (in_use == SINGLE_USER)                // Opened for O_EXCL.
   {
      if (das_ad_set_range(das_fd, io_msg.das_ad_set_range.chnl, 
                  io_msg.das_ad_set_range.range_min,
                  io_msg.das_ad_set_range.range_max) == -1)
         io_msg.status = errno;
   }
   else
   {
      if (ocb->req_t == A_D)                 // Don't allow range change if
      {                                      // A/D is in progress.
         io_msg.status = EBUSY;
         return(REPLY);
      }   
                     // Must be root or euid = 0
      get_ids(pid, &uid, &gid);
      if (uid == 0 || uid == das_stat.st_uid)
      {                 
                     // Stop A/D DMA recycling.
         das_stop_dma();
                     // Set the range.
         if (das_ad_set_range(das_fd, io_msg.das_ad_set_range.chnl, 
                        io_msg.das_ad_set_range.range_min,
                        io_msg.das_ad_set_range.range_max) == -1)
            io_msg.status = errno;
                    // Restart A/D DMA recycling.
         das_start_dma();
      }
      else
         io_msg.status = EACCES;
   }
   return(REPLY);
}

//*****************************************************************************
// das_ad_get_rng_msg() -  Get the current range of the specified channel.
//*****************************************************************************
int das_ad_get_rng_msg(pid_t pid)
{
   struct ocb *ocb;

                     // Get 'ocb' assoicated with calling process.
   if ((ocb = get_ocb(pid, io_msg.das_ad_get_range.fd)) == -1)
   {
      io_msg.status = EBADF;
      return(REPLY);
   }
                     // Check access mode & access flags
   if ((ocb->mode & S_IREAD) == 0)
   {
      io_msg.status = EPERM;
      return(REPLY);
   }
                     // Get the current range.
   if (das_ad_get_range(das_fd, io_msg.das_ad_get_range.chnl, 
                  &io_msg.das_ad_get_range_reply.range_min, 
                  &io_msg.das_ad_get_range_reply.range_max) == -1)
      io_msg.status = errno;
   else
      io_msg.status = EOK;
   return(REPLY);
}

//*****************************************************************************
// das_ad_set_trig_msg() -    Set the trigger source for A/D conversions. 
//                            (Note: This does not trigger sampling of client 
//                            requests, only the acquistion manager.) 
//                            (Note: Must be root or euid = 0, or O_EXCL).
//*****************************************************************************
int das_ad_set_trig_msg(pid_t pid)
{
   struct ocb *ocb;
   uid_t uid;
   gid_t gid;

                     // Get 'ocb' assoicated with calling process.
   if ((ocb = get_ocb(pid, io_msg.das_ad_set_trig.fd)) == -1)
   {
      io_msg.status = EBADF;
      return(REPLY);
   }
   io_msg.status = EOK;
   
   if (in_use == SINGLE_USER)                // Opened for O_EXCL
   {
      if (das_ad_set_trig(das_fd, io_msg.das_ad_set_trig.chnl, 
                        io_msg.das_ad_set_trig.trig_src) == -1)
         io_msg.status = errno;
   }
   else
   {
      if (ocb->req_t == A_D)                 // Don't allow trigger source
      {                                      // change if A/D is in progress.
         io_msg.status = EBUSY;
         return(REPLY);
      }     
                     // Must be root or euid == 0)
      get_ids(pid, &uid, &gid);
      if (uid == 0 || uid == das_stat.st_uid)
      {                      
                     // Stop A/D DMA recycleing.
         das_stop_dma();
                    // Set the trigger source.
         if (das_ad_set_trig(das_fd, io_msg.das_ad_set_trig.chnl, 
                           io_msg.das_ad_set_trig.trig_src) == -1)
            io_msg.status = errno;
                    // Restart A/D DMA recycleing.
         das_start_dma();
      }
      else
         io_msg.status = EACCES;
   }
   return(REPLY);
}

//*****************************************************************************
// das_ad_get_trig_msg() - Get the current trigger source for A/D conversions.
//*****************************************************************************
int das_ad_get_trig_msg(pid_t pid)
{
   struct ocb *ocb;

                     // Get 'ocb' assoicated with calling process.
   if ((ocb = get_ocb(pid, io_msg.das_ad_get_trig.fd)) == -1)
   {
      io_msg.status = EBADF;
      return(REPLY);
   }
                     // Check access mode & access flags
   if ((ocb->mode & S_IREAD) == 0)
   {
      io_msg.status = EACCES;
      return(REPLY);
   }
                     // Get the current trigger source.
   if (das_ad_get_trig(das_fd, io_msg.das_ad_get_trig.chnl, 
                     (int *)&io_msg.das_ad_get_trig_reply.trig_src) == -1)
      io_msg.status = errno;
   else
      io_msg.status = EOK;
   return(REPLY);
}

//*****************************************************************************
// das_ad_get_que_msg() - Get the current number of requests in A/D queues.
//*****************************************************************************
int das_ad_get_que_msg(pid_t pid)
{
   struct ocb *ocb;

                     // Get 'ocb' assoicated with calling process.
   if ((ocb = get_ocb(pid, io_msg.das_ad_get_queue.fd)) == -1)
   {
      io_msg.status = EBADF;
      return(REPLY);
   }
                     // Check access mode & access flags
   if ((ocb->mode & S_IREAD) == 0)
   {
      io_msg.status = EACCES;
      return(REPLY);
   }
                     // Request for request queue entries.
   if (io_msg.das_msg.request == _DAS_AD_GET_RDY_QUE) 
      io_msg.das_ad_get_queue_reply.n_ent = ocb->ad.n_rent;
   else              // Request for finish queue entries.
      io_msg.das_ad_get_queue_reply.n_ent = ocb->ad.n_fent;
   io_msg.status = EOK;
   return(REPLY);
}


//*****************************************************************************
// excl_das_ad_enqueue_msg() -   Retieve A/D conversions using the direct call
//                               library (for exclusive access only).
//*****************************************************************************
int excl_das_ad_enqueue_msg(struct ocb *ocb)
{
   struct das_res_queue *tmp;

   if (io_msg.das_ad_enque.mode & DAS_DMA)   // If this is a DMA request, make 
      if (gdma_flag == NULL)                 // sure data buffers are DMA-safe.
      {
         io_msg.status = EINVAL;
         return(REPLY);
      }
                     // Create que entry
   if ((tmp = das_res_create_entry()) == NULL)
   {
      io_msg.status = errno;
      return(REPLY);
   }
                        // Get data buffer.
   if ((tmp->mem_que = das_get_buf(ocb)) == NULL)
   {
      io_msg.status = ENOMEM;
      return(REPLY);
   }
                     // Number of conversions to do.  
   tmp->nconv = io_msg.das_ad_enque.n_bytes / d_info.da_res;
   tmp->fin_proxy = io_msg.das_ad_enque.proxy; // Proxy to notify client.
   tmp->mode = io_msg.das_ad_enque.mode;
   io_msg.status = EOK;
   if (das_ad_enqueue(das_fd, 0, tmp->mem_que->buffer, 
                        io_msg.das_ad_enque.n_bytes, 
                        tmp->mode, hw_adproxy) == -1)
   {
      das_rel_buf(tmp->mem_que, ocb);
      free(tmp);
      io_msg.status = errno;
      return(REPLY);
   }
   ocb->ad.n_rent++;
                     // Place entry in que.
   das_res_link(tmp, &ocb->ad.r_hd, &ocb->ad.r_tl);
   return(REPLY);
}
