//*****************************************************************************
// das_da_r.c module contains routines to process messages sent by a client to 
// perform the function of the 'das' driver routines.  The true operation of 
// the functions may not be identical to the actual low level 'das' rountine 
// because of the added feature of supporting multiple applications using the 
// same adapter.
//*****************************************************************************
// The following is the list of functions in this module.
//
//     Message processing function     DAS driver function 
//     ---------------------------     -------------------
//         das_da_sync_msg          -     das_da_sync
//         das_da_enqueue_msg       -     das_da_enqueue                       
//         das_da_wait_msg          -     das_da_wait
//                                  -     das_da_cwait            
//         das_da_term_msg          -     das_da_term
//         das_da_sam_msg           -     das_da_set_sample
//         das_da_get_rng_msg       -     das_da_get_range
//         das_da_set_rng_msg       -     das_da_set_range
//         das_da_set_trig_msg      -     das_da_set_trig
//         das_da_get_trig_msg      -     das_da_get_trig
//
//         excl_das_da_enqueue_msg  -     used for exclusive access to adapter.
//*****************************************************************************
#include <stdlib.h>
#include <malloc.h>
#include <sys/kernel.h>
#include <sys/types.h>

#include <das.h>
#include "struct.h"
#include "extern.h"
#include "proto.h"
#include "define.h"

//*****************************************************************************
// das_da_sync_msg() - Preform one D/A conversion.
//*****************************************************************************
int das_da_sync_msg(pid_t pid)
{
   struct ocb *ocb;
   das_t data;

                     // Get 'ocb' assoicated with calling process.   
   if ((ocb = get_ocb(pid, io_msg.das_da_sync.fd)) == -1)
   {
      io_msg.status = EBADF;
      return(REPLY);
   }
                     // Check access mode & access flags.
   if ((ocb->mode & S_IWRITE) == 0)
   {
      io_msg.status = EACCES;
      return(REPLY);
   }
                     // Check for exclusive access to D/A.
   if (owner_pid != 0)
   {
      io_msg.status = EBUSY;
      return(REPLY);
   }
   io_msg.status = EOK;
   
                     // Get the data from client to preform D/A conversion
   if (Readmsg(pid, sizeof(io_msg.das_da_sync) - 
               sizeof(io_msg.das_da_sync.data), &data, d_info.da_res) == -1)
      io_msg.status = errno;
      
   if (das_da_sync(das_fd, io_msg.das_da_sync.chnl, (char *)&data) == -1)
         io_msg.status = errno;
   return(REPLY);
}

//*****************************************************************************
// das_da_enqueue_msg() - Enqueue a D/A request from a client. 
//*****************************************************************************
int das_da_enqueue_msg(pid_t pid)
{
   struct ocb *ocb;
   struct das_res_queue *tmp;
   
                     // Get 'ocb' assoicated with calling process.   
   if ((ocb = get_ocb(pid, io_msg.das_da_enque.fd)) == -1)
   {
      io_msg.status = EBADF;
      return(REPLY);
   }
                     // Check access mode & access flags.
   if ((ocb->mode & S_IWRITE) == 0)
   {
      io_msg.status = EACCES;
      return(REPLY);
   }
                     // Make sure a A/D conversion is not in progress
                     // and no one else is using D/A.
   if (ocb->req_t == A_D || (owner_pid != pid && owner_pid !=0))
   {
      io_msg.status = EBUSY;
      return(REPLY);
   }
   owner_pid = pid;  
                     // Check request size with buffer size.
   if (io_msg.das_da_enque.n_bytes > gbuf_size)
   {
      io_msg.status = ENOMEM;
      return(REPLY);
   }
                     // Are we opened for O_EXCL.
   if (in_use == SINGLE_USER)             
   {
      ocb->pid = pid;
      return(excl_das_da_enque_msg(ocb));
   }
                     // Create que entry
   if ((tmp = das_res_create_entry()) == NULL)
   {
      io_msg.status = errno;
      return(REPLY);
   }
                     // Get data buffer.
   if ((tmp->mem_que = das_get_buf(ocb)) == NULL)
   {
      io_msg.status = ENOMEM;
      return(REPLY);
   }
                     // Get the data to convert from the client.
   if (Readmsg(pid, sizeof(io_msg.das_da_enque) - 
               sizeof(io_msg.das_da_enque.data), tmp->mem_que->buffer->bufp, 
               io_msg.das_da_enque.n_bytes) == -1)
   {
      io_msg.status = errno;
      return(REPLY);
   }
                     // Starting pointer of data buffer.
   tmp->cp = (das_t *)tmp->mem_que->buffer->bufp;                       
                                             // Number of conversions to do.
   tmp->nconv = io_msg.das_da_enque.n_bytes / d_info.da_res;      
   tmp->cur_conv = tmp->nconv;               // Count down for each conversion.
   tmp->fin_proxy = io_msg.das_da_enque.proxy; // Proxy to notify client.
   tmp->chnl = io_msg.das_da_enque.chnl;      // Channel to do conversions on.
   tmp->mode = io_msg.das_da_enque.mode;      
   ocb->req_t = D_A;                         // Set request type.
                     // Set queue request type if a POLL operation so that 
                     // we kown to unblock client later.
    if (io_msg.das_da_enque.mode & DAS_DMA_SYNC)
      tmp->que_req_t = POLL;
   ocb->da.n_rent++;
                     // Place entry in queue.
                     // If request is first in que, then start the conversion
   if (das_res_link(tmp, &ocb->da.r_hd, &ocb->da.r_tl))
      ocb->t_id = das_add_tick_list(ocb, ocb->da.ticks);
                     // block client if doing POLL operation.
   if (tmp->que_req_t == POLL)
      return(NO_REPLY);
   io_msg.status = EOK;
   return(REPLY);
}

//*****************************************************************************
// das_da_wait_msg() -  Get the status or wait for a previously enqueued D/A
//                      request.
//*****************************************************************************
int das_da_wait_msg(pid_t pid)
{
   struct ocb *ocb;
   
                     // Get 'ocb' assoicated with calling process.   
   if ((ocb = get_ocb(pid, io_msg.das_da_wait.fd)) == -1)
   {
      io_msg.status = EBADF;
      return(REPLY);
   }
                     // Are there any requests finished or pending?
   if (ocb->da.n_rent == 0 && ocb->da.n_fent == 0)
   {
      io_msg.status = EDEADLK;
      return(REPLY);
   }
   io_msg.status = EOK; 

                     // Is this a das_da_cwait message (non-blocking).
   if (io_msg.das_msg.request == _DAS_DA_CWAIT)
   {
      if (ocb->da.n_fent == 0)
      {
         io_msg.das_da_wait_reply.n_bytes = 0;
         return(REPLY);
      }
      Creceive(ocb->da.f_hd->fin_proxy, 0, 0);
   }                 // Has request finished yet.
   else if (ocb->da.n_fent == 0)
   {
      ocb->da.reuse = io_msg.das_da_wait.reuse;
      ocb->da.r_hd->que_req_t |= WAIT_BLK;   // Mark this request a wait blocked
      return(NO_REPLY);
   }

   if (in_use == SINGLE_USER)                // dequeue request from direct lib.
      das_da_cwait(das_fd, 0, 0);
   
   io_msg.status = EOK;
   io_msg.das_da_wait_reply.n_bytes = ocb->da.f_hd->nconv * d_info.da_res;
   ocb->da.n_fent--;
                     // Can't reuse request if client has started a A/D request.
   if (io_msg.das_da_wait.reuse == DAS_REUSE && ocb->req_t == A_D)
   {
      io_msg.status = EBUSY;
      return(REPLY);
   }
                     // Re-enqueue request.  
   if (io_msg.das_da_wait.reuse == DAS_REUSE)
   {
      if (in_use == SINGLE_USER)
      {
         ocb->da.n_rent++;
         das_da_enqueue(das_fd, 0, ocb->da.f_hd->mem_que->buffer, 
                        io_msg.das_da_enque.n_bytes,
                        ocb->da.f_hd->mode, hw_daproxy);
                     // Place entry in queue.
         das_res_link(ocb->da.f_hd, &ocb->da.r_hd, &ocb->da.r_tl);
         das_res_unlink(KEEP, &ocb->da.f_hd, &ocb->da.f_tl);      
         ocb->da.r_tl->nptr = NULL;
         return(REPLY);
      }
      if ((ocb->da.f_hd->mem_que = das_get_buf(ocb)) == NULL)
      {
         io_msg.status = ENOMEM;
         return(REPLY);
      }
                     // Starting pointer of data buffer
      ocb->da.f_hd->cp = (das_t *)ocb->da.f_hd->mem_que->buffer->bufp;
      ocb->da.f_hd->cur_conv = ocb->da.f_hd->nconv;         
                     // Place entry in queue.
                     // If request is first in queue, then start the conversions
      if (das_res_link(ocb->da.f_hd, &ocb->da.r_hd, &ocb->da.r_tl))
         ocb->t_id = das_add_tick_list(ocb, ocb->da.ticks);
      das_res_unlink(KEEP, &ocb->da.f_hd, &ocb->da.f_tl);      
      ocb->da.n_rent++;
      ocb->req_t = D_A;
   }
   else              // Unlink request from finish queue.
      das_res_unlink(FREE, &ocb->da.f_hd, &ocb->da.f_tl);
   return(REPLY);
}

//*****************************************************************************
// das_da_term_msg() -  Terminate all current, pending and finished D/A 
//                      requests.
//*****************************************************************************
int das_da_term_msg(pid_t pid)
{
   struct ocb *ocb;
   
                     // Get 'ocb' assoicated with calling process.   
   if ((ocb = get_ocb(pid, io_msg.das_da_term.fd)) == -1)
   {
      io_msg.status = EBADF;
      return(REPLY);
   }
   
   if (in_use == SINGLE_USER)                // Opened for O_EXCL.
      das_da_term(das_fd, 0);
   else 
   {
      if (ocb->req_t == D_A)                 // A/D operation in progress?
         das_rm_tick_list(ocb->t_id);        // Stop current request.
   }
                     // Clear pending requests.
   while (ocb->da.r_hd)
   {
      das_rel_buf(ocb->da.r_hd->mem_que, ocb);  
      das_res_unlink(FREE, &ocb->da.r_hd, &ocb->da.r_tl);
   }
                     // Clear finished requests.
   while (ocb->da.f_hd)
      das_res_unlink(FREE, &ocb->da.f_hd, &ocb->da.f_tl);

   ocb->req_t = NONE;
   owner_pid = 0;
   ocb->da.n_rent = 0;
   ocb->da.n_fent = 0;
   io_msg.status = EOK;
   return(REPLY);
}

//*****************************************************************************
// das_da_sam_msg() -   Set the sampling rate for D/A conversions
//*****************************************************************************
int das_da_sam_msg(pid_t pid)
{
   struct ocb *ocb;

                     // Get 'ocb' assoicated with calling process.   
   if ((ocb = get_ocb(pid, io_msg.das_da_sam.fd)) == -1)
   {
      io_msg.status = EBADF;
      return(REPLY);
   }
                     // Check access mode & access flags.
   if ((ocb->mode & S_IWRITE) == 0)
   {
      io_msg.status = EACCES;
      return(REPLY);
   }
   
   io_msg.status = EOK;   
   if (in_use == SINGLE_USER)                // Opened for O_EXCL.
   {
      if (das_da_set_sample(das_fd, io_msg.das_da_sam.chnl, 
                              io_msg.das_da_sam.ticks) == -1)
         io_msg.status = errno;
   }     
   else
   {                 // Check for valid sample rate.
      if ((io_msg.das_da_sam.ticks < MIN_SAMPLE_TICK) ||
            (io_msg.das_da_sam.ticks > MAX_SAMPLE_TICK))
      {
         io_msg.status = EINVAL;
         return(REPLY);
      }
   
      ocb->da.ticks = io_msg.das_da_sam.ticks;
      if (ocb->req_t == D_A)                 // If a request is in progress,
      {
         das_rm_tick_list(ocb->t_id);                       // Remove old entry
         ocb->t_id = das_add_tick_list(ocb, ocb->da.ticks); // Add new entry
      }
   }  
   return(REPLY);
}

//*****************************************************************************
// das_da_set_rng_msg() -  Set the current range for the specified channel 
//                         (Note: Must be root or have euid = 0, or O_EXCL).
//*****************************************************************************
int das_da_set_rng_msg(pid_t pid)
{
   struct ocb *ocb;
   uid_t uid = 0;
   gid_t gid = 0;
                     
                     // Get 'ocb' assoicated with calling process.
   if ((ocb = get_ocb(pid, io_msg.das_da_set_range.fd)) == -1)
   {
      io_msg.status = EBADF;
      return(REPLY);
   }
      
   if (in_use != SINGLE_USER)                // Not opened for O_EXCL.
   {
      if (ocb->req_t == D_A)                 // Don't allow range change if
      {                                      // D/A is in progress.
         io_msg.status = EBUSY;
         return(REPLY);
      }
      get_ids(pid, &uid, &gid);
   }
   if (uid == 0 || uid == das_stat.st_uid)   // Must be root or euid == 0.
   {
                     // Set the range
      if (das_da_set_range(das_fd, io_msg.das_da_set_range.chnl, 
                     io_msg.das_da_set_range.range_min,
                     io_msg.das_da_set_range.range_max) == -1)
         io_msg.status = errno;
      else
         io_msg.status = EOK;
   }
   else
      io_msg.status = EACCES;
   return(REPLY);
}

//*****************************************************************************
// das_da_get_rng_msg() -  Get the current range of the specified channel.
//*****************************************************************************
int das_da_get_rng_msg(pid_t pid)
{
   struct ocb *ocb;
                     
                     // Get 'ocb' assoicated with calling process.
   if ((ocb = get_ocb(pid, io_msg.das_da_get_range.fd)) == -1)
   {
      io_msg.status = EBADF;
      return(REPLY);
   }
                     // Check access mode & access flags
   if ((ocb->mode & S_IWRITE) == 0)
   {
      io_msg.status = EACCES;
      return(REPLY);
   }
                     // Get the current range.
   if (das_da_get_range(das_fd, io_msg.das_da_get_range.chnl, 
                  &io_msg.das_da_get_range_reply.range_min,
                  &io_msg.das_da_get_range_reply.range_max) == -1)
      io_msg.status = errno;
   else
      io_msg.status = EOK;
   return(REPLY);
}

//*****************************************************************************
// das_da_set_trig_msg() - Set the trigger source for D/A conversions.
//                         (Note: Must be root or euid = 0, or O_EXCL).
//*****************************************************************************
int das_da_set_trig_msg(pid_t pid)
{
   struct ocb *ocb;
   uid_t uid = 0;
   gid_t gid = 0;
                     
                     // Get 'ocb' assoicated with calling process.
   if ((ocb = get_ocb(pid, io_msg.das_da_set_trig.fd)) == -1)
   {
      io_msg.status = EBADF;
      return(REPLY);
   }
   
   if (in_use != SINGLE_USER)                // Not opened for O_EXCL.
   {
      if (ocb->req_t == D_A)                 // Don't allow trigger source
      {                                      // change is D/A is in progress
         io_msg.status = EBUSY;
         return(REPLY);
      }
      get_ids(pid, &uid, &gid);              // Must be root or euid = 0
   }
   if (uid == 0 || uid == das_stat.st_uid)
   {                 
                     // Set the trigger source.
      if (das_da_set_trig(das_fd, io_msg.das_da_set_trig.chnl, 
                        io_msg.das_da_set_trig.trig_src) == -1)
         io_msg.status = errno;
      else
         io_msg.status = EOK;
   }
   else
      io_msg.status = EACCES;
   return(REPLY);
}

//*****************************************************************************
// das_da_get_trig_msg() - Get the current trigger source for D/A conversions. 
//*****************************************************************************
int das_da_get_trig_msg(pid_t pid)
{
   struct ocb *ocb;
                     
                     // Get 'ocb' assoicated with calling process.
   if ((ocb = get_ocb(pid, io_msg.das_da_get_trig.fd)) == -1)
   {
      io_msg.status = EBADF;
      return(REPLY);
   }
                     // Check access flags.
   if ((ocb->mode & S_IWRITE) == 0)
   {
      io_msg.status = EACCES;
      return(REPLY);
   }
                     // Get the trigger source.
   if (das_da_get_trig(das_fd, io_msg.das_da_get_trig.chnl, 
                  (int *)&io_msg.das_da_get_trig_reply.trig_src) == -1)
      io_msg.status = errno;
   else
      io_msg.status = EOK;
   return(REPLY);
}

//*****************************************************************************
// das_da_get_que_msg() - Get the current number of requests in D/A queue.
//*****************************************************************************
int das_da_get_que_msg(pid_t pid)
{
   struct ocb *ocb;
                     
                     // Get 'ocb' assoicated with calling process.
   if ((ocb = get_ocb(pid, io_msg.das_da_get_queue.fd)) == -1)
   {
      io_msg.status = EBADF;
      return(REPLY);
   }
                     // Check access mode & access flags
   if ((ocb->mode & S_IWRITE) == 0)
   {
      io_msg.status = EACCES;
      return(REPLY);
   }
                     // Request for request queue entries.   
   if (io_msg.das_msg.request == _DAS_DA_GET_RDY_QUE)
      io_msg.das_da_get_queue_reply.n_ent = ocb->da.n_rent;
   else              // Request for finish queueu entries.
      io_msg.das_da_get_queue_reply.n_ent = ocb->da.n_fent;
      
   io_msg.status = EOK;
   return(REPLY);
}

//*****************************************************************************
// excl_das_da_int_dma_msg() - 
//*****************************************************************************
int excl_das_da_enque_msg(struct ocb *ocb)
{
   struct das_res_queue *tmp;

   if (io_msg.das_ad_enque.mode & DAS_DMA)   // If this is a DMA request, make
      if (gdma_flag == NULL)                 // sure data buffers are DMA-safe.
      {
         io_msg.status = EINVAL;
         return(REPLY);
      }
                     // Create que entry
   if ((tmp = das_res_create_entry()) == NULL)
   {
      io_msg.status = errno;
      return(REPLY);
   }
                     // Get data buffer.
   if ((tmp->mem_que = das_get_buf(ocb)) == NULL)
   {
      io_msg.status = ENOMEM;
      return(REPLY);
   }
                     // Get the data to convert from the client.
   Readmsg(ocb->pid, sizeof(io_msg.das_da_enque) - 
            sizeof(io_msg.das_da_enque.data), tmp->mem_que->buffer->bufp, 
            io_msg.das_da_enque.n_bytes);
   
   tmp->mode = io_msg.das_ad_enque.mode;
                     // Number of conversions to do.
   tmp->nconv = io_msg.das_da_enque.n_bytes * d_info.da_res;
   tmp->fin_proxy = io_msg.das_da_enque.proxy; // Proxy to notify client.
   io_msg.status = EOK;
   if (das_da_enqueue(das_fd, 0, tmp->mem_que->buffer, 
                        io_msg.das_da_enque.n_bytes, tmp->mode, 
                        hw_daproxy) == -1)
   {
      das_rel_buf(tmp->mem_que, ocb);
      free(tmp);
      io_msg.status = errno;
      return(REPLY);
   }
   ocb->da.n_rent++;
                     // Place entry in que.
   das_res_link(tmp, &ocb->da.r_hd, &ocb->da.r_tl);
   return(REPLY);
}
