//*****************************************************************************
// das_init.c module contain the initialization and termination routines used 
// in the acquistion manager.  
//*****************************************************************************
// The following is the list of functions in this module.
//
//    -  terminate   
//    -  init
//    -  das_start_dma
//    -  das_stop_dma
//    -  das_reinit
//*****************************************************************************
#include <stdlib.h>
#include <signal.h>
#include <stddef.h>
#include <unistd.h>
#include <string.h>
#include <sys/psinfo.h>
#include <sys/proxy.h>
#include <sys/osinfo.h>
#include <sys/sched.h>
#include <sys/prfx.h>
#include <sys/fd.h>

#include <das.h>
#include "struct.h"
#include "extern.h"
#include "proto.h"

int devno;
char prefix[_MAX_PATH] = {"/dev/"};

//*****************************************************************************
// terminate   -  This function is called when a terminate or kill signal is 
// sent to the acquistion manager.  This function will detach the device and the
// prefix, and close down the 'das' low level driver and free its memory.    
//*****************************************************************************
void terminate()
{
   qnx_device_detach(devno);
   qnx_prefix_detach(prefix);
   das_ad_term(das_fd, 0);
   das_da_term(das_fd, 0);
   das_free(das_fd, das_ad_data);
   das_free(das_fd, das_mem_bufs);
   das_close(das_fd);
   exit(EXIT_SUCCESS);
}

//*****************************************************************************
// das_init - This function is called from the main() during the initialization 
// of the Acquisition Manager.  This function will register this application as 
// a device driver, notify the kernel that it is a server so that it will 
// receive system messages and open the 'das' low level driver.
//*****************************************************************************
void das_init(int argc, char *argv[])
{
   int i, opt;
   struct _osinfo osdata;
   char cfg_file[_MAX_PATH] ={0};
   struct das_mem_que *f_list_ptr;
   
                     // Setup default command line parameters.
   sample_r =  10000;
   gbuf_size = 1000;
   gnum_bufs = 10;
   gdma_flag = NULL;
   
                    // Parse command line for arguments.
   while((opt = getopt(argc, argv, "P:s:S:b:f:d")) != -1)
   {
      switch(opt)
      {
         case 'P':   // Prefix
            strcat(prefix, optarg);
            break;
         
         case 's':   // Sample rate.
            sample_r = atol(optarg);
            break;
            
         case 'f':   // Configuration filename.
            strcpy(cfg_file, optarg);
            break;
         
         case 'b':   // Number of buffers to allocate.
            gnum_bufs = atoi(optarg);
            break;
            
         case 'S':   // Size of each buffer to allocate.
            gbuf_size = atoi(optarg);
            break;   

         case 'd':
            gdma_flag = _PMF_DMA_SAFE;
            break;
      }
   }
   if (strcmp("/dev/", prefix) == 0)         
      strcat(prefix, "das");                 // Set defualt prefix.
      
                     // Get system messages.
   if (qnx_pflags(~0, _PPF_SERVER, NULL, NULL)) 
   {
      perror("\npflags");
      exit(EXIT_FAILURE);
   }
                     // Attach a device number to the acquistion manager.
   if ((devno = qnx_device_attach()) == -1) 
   {
      perror("\nqnx_device_attach");
      exit(EXIT_FAILURE);
   }

   if (qnx_osinfo(0, &osdata) == -1)       
   {
      perror("\nqnx_osinfo");
      exit(EXIT_FAILURE);
   }  
                     // Get a pointer to system time segment.
   timep = MK_FP(osdata.timesel, offsetof(struct _timesel, seconds));

                     // Attach directory prefix.
   if (qnx_prefix_attach(prefix, NULL, 1))  
   {
      perror("\nqnx_prefix_attach");
      exit(EXIT_FAILURE);
   }

   if (!(fd_ctrl = __init_fd(getpid())))
   {
      perror("\n__init_fd");
      exit(EXIT_FAILURE);
   }
                     // Initialize das_stat structure.
                                       // Major and Minor Device number.
   das_stat.st_rdev = (dev_t) ((getnid() << 16L) + (devno << 10L)); 
   das_stat.st_mode = 0600 | S_IFCHR; // Permissions and file type.
   das_stat.st_nlink = 1;             // number of links.
   das_stat.st_ftime = *timep;        // Time of creation.
   das_stat.st_atime = *timep;        // Time of last access.
   das_stat.st_mtime = *timep;        // Time of last modification.
   das_stat.st_ctime = *timep;        // Time of last status change.
   das_stat.st_uid = 0;
   das_stat.st_gid = 0;
   das_stat.st_size = 0;
   
   setprio(0, 19);   // Boost our priority.

                     // Attach a proxy for A/D event notification.
   if ((hw_adproxy = qnx_proxy_attach(0, 0, 0, -1)) == -1)
   {
      perror("\nqnx_proxy_attach - A/D");
      exit(EXIT_FAILURE);
   }
                     // Attach a proxy for D/A event notification.                  
   if ((hw_daproxy = qnx_proxy_attach(0, 0, 0, -1)) == -1)
   {
      perror("\nqnx_proxy_attach - D/A");
      exit(EXIT_FAILURE);
   }
                     // Open low level 'das' driver.
   if ((das_fd = das_open(NULL, NULL, 
                           (cfg_file[0] == NULL) ? NULL : cfg_file)) == -1)
   {
      perror("\nCould not open low level driver");
      print_usage(argv);
      exit(EXIT_FAILURE);
   }
                     // Get information on adapter.
   das_get_info(das_fd, &d_info);

                     // Determine number of channel to scan.
   num_ai = (d_info.chnl_cnfg == DAS_DIFFERENTIAL) ? 8 : 16; 
   
                     // Check tick/sample 
   if ((sample_r > MAX_SAMPLE_TICK/num_ai) || 
         (sample_r < MIN_SAMPLE_TICK/num_ai))
   {
      das_close(das_fd);
      fprintf(stderr, "Invalid sample rate\n");
      print_usage(argv);
      exit(EXIT_FAILURE);
   }
   
   if (das_ad_set_sample(das_fd, 0, sample_r/num_ai) == -1)
   {                                   
      perror("\nError setting sample rate!");
      das_close(das_fd);
      exit(EXIT_FAILURE);
   }
   
   // Allow only SIGTERM & SIGKILL for graceful termination, ignore all others.
   for ( i = _SIGMIN; i <= _SIGMAX; ++i) 
      if ((i != SIGTERM) || (i != SIGKILL))
         signal( i, SIG_IGN);

   signal( SIGTERM, &terminate );
   signal( SIGKILL, &terminate );
   
                     // Allocate global data buffers
   if ((das_mem_bufs = das_alloc(das_fd, gnum_bufs, gbuf_size, 
                                 gdma_flag)) == NULL)
   {
      perror("\nCan't 'das_alloc' global memory buffers");
      exit(EXIT_FAILURE);
   }
   
   if ((glb_free_list = malloc(sizeof(struct das_mem_que) * gnum_bufs)) == NULL)
   {
      perror("\nCan't 'alloc' global memory list");
      exit(EXIT_FAILURE);
   }     
                     // Place buffers on a free list.
   f_list_ptr = glb_free_list;
   for (i = 0; i < gnum_bufs-1; i++, f_list_ptr++)
   {
      f_list_ptr->buffer = &das_mem_bufs[i];
      f_list_ptr->n_que = f_list_ptr+1;
   }
   f_list_ptr->buffer = &das_mem_bufs[i];
   f_list_ptr->n_que = NULL;
   
                     // Allocate active sample buffer (DMA safe).
   if ((das_ad_data = das_alloc(das_fd, 1, num_ai * d_info.ad_res, 
                                 _PMF_DMA_SAFE)) == NULL)
   {
      perror("\nCan't 'das_alloc' DMA safe memory");
      exit(EXIT_FAILURE);
   }

   tickque = 0;
   q[tickque].qnext = q[tickque].qprev = 0;
   q[tickque].qticks.small = -1;
   q[tickque].qticks.big = -1;
   qticks.small = 0;                         // Init global ticks
   qticks.big = 0;
}

//*****************************************************************************
// das_start_dma  -  Start A/D conversions using DAS_DMA / DAS_NOTIFY
//*****************************************************************************
void das_start_dma(void)
{
                     // Start DMA recycle of A/D conversions.
   if (das_ad_enqueue(das_fd, 0, das_ad_data, num_ai * d_info.ad_res, 
                        DAS_DMA | DAS_NOTIFY, hw_adproxy) == -1)
   {
      perror("\ndas_ad_enqueue");
      das_free(das_fd, das_mem_bufs);
      das_close(das_fd);
      exit(EXIT_FAILURE);
   }
}

//*****************************************************************************
// das_stop_dma   -  Stop A/D conversion.
//*****************************************************************************
void das_stop_dma(void)
{
   das_ad_term(das_fd, 0);
}     

//*****************************************************************************
// das_reinit  -  Reset direct library after O_EXCL 
//*****************************************************************************
void das_reinit(void)
{
   static char scan_list[] = {0, 1, 2, 3, 4, 5, 6, 7,
                             8, 9,10,11,12,13,14,15};

                     // Setup scan queue.
   das_ad_set_scan(das_fd, scan_list, num_ai);
   
                     // Set sample rate for analog input. A max of 
                     // 5000 sam/sec is allowed. 
   das_ad_set_sample(das_fd, 0, sample_r/num_ai);
}
