//*****************************************************************************
// das_proxy.c module contains the support routines to perform timed 
// data acquisition for both A/D and D/A conversions.  These routines are call 
// when a proxy is trigger by the DAS library interrupt routine to indicate that
// a conversion is to be preformed.  Also the proxy handlers for exclusive 
// access are also included.
//*****************************************************************************
// The following is the list of functions in this module.
//
//       -  das_proxy
//       -  das_put_da_sample
//       -  das_get_ad_sample
//    
//       -  excl_da_proxy
//       -  excl_ad_proxy
//*****************************************************************************
#include <stdlib.h>
#include <sys/kernel.h>
#include <das.h>
#include "struct.h"
#include "define.h"
#include "extern.h"
#include "proto.h"

//*****************************************************************************
// das_put_da_sample -   Preform a timed D/A operation.
//*****************************************************************************
void das_put_da_sample(struct ocb *ocb)
{
   int chnl;
   register struct das_res_queue *qptr = ocb->da.r_hd;
   
                     // Get 'ocb' assoicated with calling process.   
   chnl = qptr->chnl; 
                     // Perform D/A conversion.
   das_da_sync(das_fd, chnl, (char *)qptr->cp);

   qptr->cp++;                               // Update data pointer.
   ocb->da.r_hd->cur_conv--;                 // Update # of conversions.
   if (qptr->mode & DAS_DUAL_CHNL) 
   {                                         // Preform dual channel operation.
      das_da_sync(das_fd, (chnl ? 0 : 1), (char *)qptr->cp);
      qptr->cp++;
      qptr->cur_conv--;
   }
   if (qptr->cur_conv == 0)          // Are conversions done?
   {
      if (qptr->mode & DAS_RECYCLE) // Are we in recycle mode
      {
         qptr->cur_conv = qptr->nconv;
         qptr->cp = (das_t *)qptr->mem_que->buffer->bufp;   
         ocb->t_id = das_add_tick_list(ocb, ocb->da.ticks);
         return;
      }
      ocb->da.n_rent--;
                     // Is the client blocked on the manager.
      if (qptr->que_req_t & (POLL | WAIT_BLK))
      {
         io_msg.status = EOK;
         io_msg.das_da_wait_reply.n_bytes = ocb->da.r_hd->nconv * d_info.da_res;
         Reply(ocb->pid, &io_msg, sizeof(io_msg));
         if (ocb->da.reuse)               // Does the client want this request
         {                                // reused.
            qptr->cp = (das_t *)qptr->mem_que->buffer->bufp; 
            qptr->cur_conv = qptr->nconv; 
            qptr->que_req_t &= ~WAIT_BLK;
            ocb->da.n_rent++;
            ocb->da.reuse = 0;
            das_res_link(qptr, &ocb->da.r_hd, &ocb->da.r_tl);
            das_res_unlink(KEEP, &ocb->da.r_hd, &ocb->da.r_tl);
            ocb->da.r_tl->nptr = NULL;
                     // Add to time queue.
            ocb->t_id = das_add_tick_list(ocb, ocb->da.ticks); 
            return;
         }
      }
      Trigger(qptr->fin_proxy);      // Notify client that it is done.
      das_rel_buf(qptr->mem_que, ocb);  
      das_res_link(qptr, &ocb->da.f_hd, &ocb->da.f_tl);
      das_res_unlink(KEEP, &ocb->da.r_hd, &ocb->da.r_tl);
      ocb->da.f_tl->nptr = NULL;
      if (ocb->da.r_hd == NULL)        // Are there anymore requests?
      {
         owner_pid = 0;                  
         ocb->req_t = NONE;
      }
      else
         ocb->t_id = das_add_tick_list(ocb, ocb->da.ticks);       
      ocb->da.n_fent++;
      return;
   }
   ocb->t_id = das_add_tick_list(ocb, ocb->da.ticks);
   return;
}

//*****************************************************************************
// das_get_ad_sample - Perform a timed A/D operation.
//*****************************************************************************
void das_get_ad_sample(struct ocb *ocb)
{
   struct _mxfer_entry mx[2];
   register struct das_res_queue *qptr = ocb->ad.r_hd;

                     // Get the data from DMA buffer.
   *qptr->cp = das_res_buf[qptr->chnl];
   qptr->cp++;                         // Update data pointer.
   if (ocb->chnl_ent == ocb->ad_seq.n_entries-1)
      ocb->chnl_ent = 0;                     // Reset scan que.
   else
      ocb->chnl_ent++;                       // Update scan que pointer.
   qptr->chnl = ocb->ad_seq.seq[ocb->chnl_ent];
   if ((--qptr->cur_conv) == 0)      // Are conversions done?
   {
      ocb->ad.n_rent--;
      if (qptr->que_req_t & (READ_BLK | POLL))
      {
                     // Setup reply header.
         io_msg.das_ad_read_reply.status = EOK;
         io_msg.das_ad_read_reply.n_bytes = qptr->nconv * d_info.ad_res;
         _setmx(&mx[0], &io_msg, sizeof(io_msg.das_ad_read_reply) - 
                  sizeof(io_msg.das_ad_read_reply.data));
                     // Setup data buffer to send.
         _setmx(&mx[1], qptr->mem_que->buffer->bufp, 
                  io_msg.das_ad_read_reply.n_bytes);
         Replymx(ocb->pid, 2, &mx);          // Reply with header and data.
         if (ocb->ad.reuse)
         {
            qptr->cp = (das_t *)qptr->mem_que->buffer->bufp;               
            qptr->cur_conv = qptr->nconv;     // Count down for each conversion.
            qptr->chnl = ocb->ad_seq.seq[0];  // Get first chnl in scan queue.
            qptr->que_req_t &= ~READ_BLK;
            ocb->req_t = A_D;                // Set request type.
            ocb->ad.n_rent++;
            ocb->ad.reuse = NULL;
            das_res_link(qptr, &ocb->ad.r_hd, &ocb->ad.r_tl);
            das_res_unlink(KEEP, &ocb->ad.r_hd, &ocb->ad.r_tl);
            ocb->ad.r_tl->nptr = NULL;
            ocb->chnl_ent = 0;                     // Scan que entry #.
                     // Add to time queue.
            ocb->t_id = das_add_tick_list(ocb, ocb->ad.ticks); 
            return;
         }
         das_rel_buf(qptr->mem_que, ocb);
                     // Remove from queue.
         das_res_unlink(FREE, &ocb->ad.r_hd, &ocb->ad.r_tl);
         ocb->chnl_ent = 0;                  // Reset scan que pointer.
         if (ocb->ad.r_hd == NULL)
            ocb->req_t = NONE;
         else
            ocb->t_id = das_add_tick_list(ocb, ocb->ad.ticks);       
         return;
      }
      Trigger(qptr->fin_proxy);     // Notify client that it is done.  
      das_res_link(qptr, &ocb->ad.f_hd, &ocb->ad.f_tl);
      das_res_unlink(KEEP, &ocb->ad.r_hd, &ocb->ad.r_tl);
      ocb->ad.f_tl->nptr = NULL;
      ocb->chnl_ent = 0;                  // Reset scan que pointer.
      if (ocb->ad.r_hd == NULL)
         ocb->req_t = NONE;
      else
         ocb->t_id = das_add_tick_list(ocb, ocb->ad.ticks);       
      ocb->ad.n_fent++;
      return;   
   }
   ocb ->t_id = das_add_tick_list(ocb, ocb->ad.ticks);
   return;
}

//*****************************************************************************
// excl_da_proxy_msg -  Process exclusive access request proxies for D/A 
//                      requests.
//*****************************************************************************
void excl_da_proxy_msg()
{
   if (excl_ocb->da.r_hd->que_req_t & WAIT_BLK) // Is client BLOCKED on manager.
   {
      das_da_cwait(das_fd, 0, 0);
      io_msg.status = EOK;
                     // Setup reply header.
      io_msg.das_ad_read_reply.n_bytes = excl_ocb->ad.r_hd->nconv * 
                                          d_info.ad_res;
      Reply(excl_ocb->pid, &io_msg, sizeof(io_msg));
      if (excl_ocb->ad.reuse)
      {
         das_ad_enqueue(das_fd, 0, excl_ocb->ad.r_hd->mem_que->buffer, 
                           io_msg.das_ad_read_reply.n_bytes, 
                           excl_ocb->ad.r_hd->mode, hw_adproxy);
                     // Place entry in que.
         das_res_link(excl_ocb->ad.r_hd, &excl_ocb->ad.r_hd, &excl_ocb->ad.r_tl);
         das_res_unlink(KEEP, &excl_ocb->ad.r_hd, &excl_ocb->ad.r_tl);
         excl_ocb->ad.r_tl->nptr = NULL;
         return;
      }
      das_rel_buf(excl_ocb->da.r_hd->mem_que, excl_ocb);
      das_res_unlink(FREE, &excl_ocb->da.r_hd, &excl_ocb->da.r_tl);
      excl_ocb->da.n_rent--;
      return;
   }
   Trigger(excl_ocb->da.r_hd->fin_proxy);       // Notify client that it is done.  
   das_res_link(excl_ocb->da.r_hd, &excl_ocb->da.f_hd, &excl_ocb->da.f_tl);
   das_res_unlink(KEEP, &excl_ocb->da.r_hd, &excl_ocb->da.r_tl);
   excl_ocb->da.f_tl->nptr = NULL;  
   excl_ocb->da.n_rent--;
   excl_ocb->da.n_fent++;
}

//*****************************************************************************
// excl_ad_proxy_msg -  Process exclusive access request proxies for A/D 
//                      requests.
//*****************************************************************************
void excl_ad_proxy_msg()
{
   struct _mxfer_entry  mx[2];
   
   if (excl_ocb->ad.r_hd->que_req_t & READ_BLK)
   {
      das_ad_dequeue(das_fd, 0);
      io_msg.status = EOK;
                     // Setup reply header.
      io_msg.das_ad_read_reply.n_bytes = excl_ocb->ad.r_hd->nconv * 
                                             d_info.ad_res;
      _setmx(&mx[0], &io_msg, sizeof(io_msg.das_ad_read_reply) - 
               sizeof(io_msg.das_ad_read_reply.data));
                     // Setup data buffer to send.
      _setmx(&mx[1], excl_ocb->ad.r_hd->mem_que->buffer->bufp, 
               io_msg.das_ad_read_reply.n_bytes);
      Replymx(excl_ocb->pid, 2, &mx);  // Reply with header and data.
      if (excl_ocb->ad.reuse)
      {
         das_ad_enqueue(das_fd, 0, excl_ocb->ad.r_hd->mem_que->buffer, 
                           io_msg.das_ad_read_reply.n_bytes, 
                           excl_ocb->ad.r_hd->mode, hw_adproxy);
                     // Place entry in que.
         das_res_link(excl_ocb->ad.r_hd, &excl_ocb->ad.r_hd, &excl_ocb->ad.r_tl);
         das_res_unlink(KEEP, &excl_ocb->ad.r_hd, &excl_ocb->ad.r_tl);
         excl_ocb->ad.r_tl->nptr = NULL;
         return;
      }
      das_rel_buf(excl_ocb->ad.r_hd->mem_que, excl_ocb);
      das_res_unlink(FREE, &excl_ocb->ad.r_hd, &excl_ocb->ad.r_tl);
      excl_ocb->ad.n_rent--;
      return;
   }
   Trigger(excl_ocb->ad.r_hd->fin_proxy);   // Notify client that it is done.  
   das_res_link(excl_ocb->ad.r_hd, &excl_ocb->ad.f_hd, &excl_ocb->ad.f_tl);
   das_res_unlink(KEEP, &excl_ocb->ad.r_hd, &excl_ocb->ad.r_tl);
   excl_ocb->ad.f_tl->nptr = NULL;  
   excl_ocb->ad.n_rent--;
   excl_ocb->ad.n_fent++;
}
