/* Demonstration program that "streams" A/D conversions to disk until a key
   is pressed.  Note the value of 'N_ENT', 'N_LOW' and 'N_HIGH' may need to
   be increased if you are logging at high rates to a remote disk.
   Note:  This program will only work with the direct call library.
*/

#include <stdio.h>
#include <stdlib.h>
#include <conio.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/proxy.h>
#include <sys/kernel.h>
#include <das.h>

#define N_ENT        10             // number of entries
#define N_LOW        5              // min number of entries in request queue
#define N_HIGH       5              // max number of entries in finish queue
#define N_CONV       32000          // 5000 conversions
#define C_RATE       100000         // 100kHz conversion rate
   // convert number of conversions to bytes
#define get_bytes(num) (num * dinfo.ad_res)

void stream(void);

int fd, fp;
das_buf_t   *bufp;
das_info    dinfo;

void main(void)
{
   char slist[] = { 0, 1, 2, 3, 4, 5, 6, 7 };         /* scan sequence */

   /* open the 'das' resource manager/library               */
   if ((fd = das_open("/dev/das", O_RDWR|O_EXCL, 0)) == -1)
   {
      perror("das_open");
      exit(EXIT_FAILURE);
   }

   /* open a file to log the data to                        */
   if ((fp = open("das.dat", O_WRONLY | O_CREAT | O_TRUNC, S_IRUSR | S_IWUSR)) == -1)
   {
      perror("file open");
      exit(EXIT_FAILURE);
   }

   /* get hardware specific information on adapter          */
   if(das_get_info(fd, &dinfo) == -1)
   {
      perror("das_get_info");
      exit(EXIT_FAILURE);
   }

   /* set the conversion rate on channel 0                  */
   if (das_ad_set_sample(fd, 0, dinfo.pace_clk / C_RATE) == -1)
   {
      perror("das_ad_set_sample");
      exit(EXIT_FAILURE);
   }

   /* set the trigger source on to the onboard pacer/timer  */
   if (das_ad_set_trig(fd, 0, DAS_PC_TRIG) == -1)
   {
      perror("das_ad_set_trig");
      exit(EXIT_FAILURE);
   }

   /* set the range to -5V - +5V                            */
   if (das_ad_set_range(fd, 0, -5000000, 5000000) == -1)
   {
      perror("das_ad_set_range");
      exit(EXIT_FAILURE);
   }
      
   /* set a scan sequence 0 .. 7                            */
   if (das_ad_set_scan(fd, slist, sizeof(slist)) == -1)
   {
      perror("das_set_scan");
      exit(EXIT_FAILURE);
   }

   stream();             /* capture the data                */
   das_close(fd);
}

/* log the data to disk                                     */
void stream()
{
   int cnt, n_rent, wp=0, rp=0;
   pid_t proxy;
   
   printf("Logging Data\n");
   
   /* allocate DMA safe memory         */
   if ((bufp = das_alloc(fd, N_ENT, get_bytes(N_CONV), _PMF_DMA_SAFE)) == NULL)
   {
      perror("das_alloc");
      exit(EXIT_FAILURE);
   }  
   if((proxy = qnx_proxy_attach(0, 0, 0, -1)) == -1)
   {
      perror("qnx_proxy_attach");
      exit(EXIT_FAILURE);
   }
   
   for(cnt=0; cnt < N_ENT; cnt++)   /* initiate 'N_ENT' requests        */
      if (das_ad_enqueue(fd, 0, &bufp[cnt], get_bytes(N_CONV), DAS_DMA | DAS_SNGL_CYCLE, proxy) == -1)
      {
         perror("das_ad_enqueue");
         das_ad_term(fd, 0);
         exit(EXIT_FAILURE);
      }

   while(!kbhit())               /* log data until a key is pressed     */
   {
      if(Receive(proxy, 0, 0) == -1)/* wait for a request to complete   */
      {
         perror("Receive");
         das_ad_term(fd, 0);
         exit(EXIT_FAILURE);
      }
      if(das_ad_dequeue(fd, 0) == NULL)   /* dequeue the entry          */
      {
         perror("das_ad_dequeue");
         das_ad_term(fd, 0);
         exit(EXIT_FAILURE);
      }
      printf("Writing buffer %d to disk\n", wp);
      if(write(fp, bufp[wp++].bufp, get_bytes(N_CONV)) < get_bytes(N_CONV))
      {
         perror("write");
         das_ad_term(fd, 0);
         exit(EXIT_FAILURE);
      }
      if(wp == N_ENT)
         wp = 0;

      /* don't let finish entries go above 50% */
      if(das_ad_fqueue(fd, 0) > N_HIGH)
      {
         for(cnt=0; cnt < N_HIGH; cnt++)
         {
            if(Receive(proxy, 0, 0) == -1)/* wait for a request to complete */
            {
               perror("Receive");
               das_ad_term(fd, 0);
               exit(EXIT_FAILURE);
            }
            if(das_ad_dequeue(fd, 0) == NULL)   /* dequeue the entry    */
            {
               perror("das_ad_dequeue");
               das_ad_term(fd, 0);
               exit(EXIT_FAILURE);
            }
            
            printf("Writing buffer %d to disk\n", wp);
            if(write(fp, bufp[wp++].bufp, get_bytes(N_CONV)) < get_bytes(N_CONV))
            {
               perror("write");
               das_ad_term(fd, 0);
               exit(EXIT_FAILURE);
            }
            if(wp == N_ENT)
               wp = 0;
         }
      }
      
      /* make sure there is always requests in the queue */
      if((n_rent = das_ad_rqueue(fd, 0)) == 0)
      {
         printf("Data has been lost!\n");
         das_ad_term(fd, 0);
         exit(-1);
      }

      if(n_rent < N_LOW)         // don't let request entries drop below 50%
      {
         for(cnt=0; (cnt < n_rent) && (rp != wp); cnt++)
         {
            printf("requesting buffer %d\n", rp);
            if (das_ad_enqueue(fd, 0, &bufp[rp++], get_bytes(N_CONV), DAS_DMA | DAS_SNGL_CYCLE, proxy) == -1)
            {
               perror("das_ad_enqueue");
               das_ad_term(fd, 0);
               exit(EXIT_FAILURE);
            }
            if(rp == N_ENT)
               rp = 0;
         }
      }
   }

   printf("Terminating all requests\n");
   if (das_ad_term(fd, 0) == -1)             /* terminate conversions   */
   {
      perror("das_ad_term");
      exit(EXIT_FAILURE);
   }

   das_free(fd, bufp);                       /* free memory             */
   qnx_proxy_detach(proxy);
   exit(EXIT_SUCCESS);
}
