/* 
   Demonstration program that "streams" A/D conversions to disk until a key
   is pressed.  Note the value of 'N_ENT', 'NUM_CONV' may need to be increased
   if you are logging at high rates to a remote disk.
*/

#include <stdio.h>
#include <stdlib.h>
#include <conio.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/proxy.h>
#include <sys/kernel.h>
#include <das.h>

#define N_ENT        10             // number of entries
#define NUM_CONV     4000           // number of conversions
#define FREQUENCY    100000L
   // convert number of conversions to bytes
#define get_bytes(num) (num * dinfo.ad_res)

void stream(void);

int fd, fp;
das_info dinfo;

void main(void)
{
   static char slist[] = { 0, 1, 2, 3, 4, 5, 6, 7 };     /* scan sequence */

   /* open the 'das' acquisition manager/library               */
   if ((fd = das_open("/dev/das", O_RDWR|O_EXCL, NULL)) == -1)
   {
      perror("das_open");
      exit(EXIT_FAILURE);
   }

   /* open a file to log the data to                        */
   if ((fp = open("das.dat", O_WRONLY | O_CREAT | O_TRUNC, S_IRUSR | S_IWUSR)) == -1)
   {
      perror("file open");
      exit(EXIT_FAILURE);
   }
   
   /* get hardware specific information on adapter          */
   if(das_get_info(fd, &dinfo) == -1)
   {
      perror("das_get_info");
      exit(EXIT_FAILURE);
   }

   /* set the conversion rate to 100 clock ticks on channel 0 */
   if (das_ad_set_sample(fd, 0, dinfo.pace_clk / FREQUENCY) == -1)
   {
      perror("das_ad_set_sample");
      exit(EXIT_FAILURE);
   }

   /* set the trigger source on to the onboard pacer/timer  */
   if (das_ad_set_trig(fd, 0, DAS_PC_TRIG) == -1)
   {
      perror("das_ad_set_trig");
      exit(EXIT_FAILURE);
   }

   /* set the range to -5V - +5V                            */
   if (das_ad_set_range(fd, 0, -5000000, 5000000) == -1)
   {
      perror("das_ad_set_range");
      exit(EXIT_FAILURE);
   }
      
   /* set a scan sequence 0 .. 7                            */
   if (das_ad_set_scan(fd, slist, sizeof(slist)) == -1)
   {
      perror("das_set_scan");
      exit(EXIT_FAILURE);
   }

   stream();             /* capture the data                */

   if(das_close(fd) == -1)
   {
      perror("das_close");
      exit(EXIT_FAILURE);
   }
   exit(EXIT_SUCCESS);
}

/* log the data to disk                                     */
void stream()
{
   das_buf_t *buf;
   char *cbuf, cnt;
   long bytes = 0;
   pid_t proxy;
   
   printf("Logging Data\n");
   
   if((proxy = qnx_proxy_attach(0, 0, 0, -1)) == -1)
   {
      perror("qnx_proxy_attach");
      exit(EXIT_FAILURE);
   }

   if ((buf = das_alloc(fd, N_ENT, get_bytes(NUM_CONV), _PMF_DMA_SAFE)) == NULL)
   {
      perror("das_alloc");
      exit(EXIT_FAILURE);
   }
   
   if((cbuf = (char *)malloc(get_bytes(NUM_CONV))) == NULL)
   {
      perror("malloc");
      exit(EXIT_FAILURE);
   }
   
   for(cnt=0; cnt < N_ENT; cnt++)   /* initiate 'N_ENT' requests        */
      if (das_ad_enqueue(fd, 0, &buf[cnt], get_bytes(NUM_CONV), DAS_SNGL_CYCLE | DAS_DMA, proxy) == -1)
      {
         perror("das_ad_enqueue");
         das_ad_term(fd, 0);
         exit(EXIT_FAILURE);
      }

   while(!kbhit())               /* log data until a key is pressed  */
   {
      Receive(proxy, 0, 0);
      /* wait for a request to complete and reuse the same buffer    */
      if(das_ad_read(fd, cbuf, get_bytes(NUM_CONV), 0, DAS_REUSE) == -1)
      {
         perror("das_ad_read");
         das_ad_term(fd, 0);
         exit(EXIT_FAILURE);
      }
      bytes+= get_bytes(NUM_CONV);
      printf("Writing buffer to disk.  Bytes so far: %ld\n", bytes);
      if(write(fp, cbuf, get_bytes(NUM_CONV)) < get_bytes(NUM_CONV))
      {
         perror("write");
         das_ad_term(fd, 0);
         exit(EXIT_FAILURE);
      }

      /* make sure there is always requests in the queue */
      if(das_ad_rqueue(fd, 0) == 0)
      {
         printf("Data has been lost!\n");
         das_ad_term(fd, 0);
         exit(-1);
      }
   }

   printf("Terminating all requests\n");
   if (das_ad_term(fd, 0) == -1)             /* terminate conversions   */
   {
      perror("das_ad_term");
      exit(EXIT_FAILURE);
   }

   das_free(fd, buf);                        /* free memory             */
   qnx_proxy_detach(proxy);
}
