// !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
// das.h contains defines and function prototypes for the data
// acquisition library.
// !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

#ifndef _DAS_HDR_H_INCLUDED
#include <das_types.h>
#include <sys/seginfo.h>

// Request types for das_ad_enqueue and das_da_enqueue functions
#define DAS_INT            0x100
#define DAS_DMA            0x200
#define DAS_DMA_SYNC       0x400

// Modes for A/D and D/A
#define DAS_SNGL_CYCLE     0x01     // perform 'N' conversions and stop 
#define DAS_RECYCLE        0x02     // conversions are repeated continuously
                                    // to the same memory.
#define DAS_SNGL_CHNL      0x04     // perform D/A conversions on single chnl
#define DAS_DUAL_CHNL      0x08     // perform D/A conversions on two chnls
#define DAS_NOTIFY         0x10     // use by Resource Manager only

#define DAS_REUSE          0x01     // reuse the same buffer

// Specific to the pcl-818
// (gain/range can be set with software (REMOTE) or by switches (LOCAL))
#define DAS_LOCAL          0x00     
#define DAS_REMOTE         0x01

// Analog input types
#define DAS_SINGLE_ENDED   0x01     // single-ended input (16 channels)
#define DAS_DIFFERENTIAL   0x00     // differential input (8 channels)

// Triggering sources
#define DAS_SW_TRIG        0x01     // software trigger
#define DAS_EX_TRIG        0x02     // external trigger
#define DAS_PC_TRIG        0x04     // pacer (counter/timer) trigger
// gateing is enabled by bitwise OR-ing ENB_GATE with a trigger source
#define DAS_ENB_GATE       0x08     // gating enable
#define DAS_A_LOW          0x10     // active low level gating
#define DAS_A_HIGH         0x20     // active high level gating

   // Function prototypes

// Digital I/O Routines
extern int das_digital_in(int fd, int port, long *bits);
extern int das_digital_out(int fd, int port, long bits, long mask, long *old_bits, long *new_bits);

// Analog to Digital Conversion Routines

extern int das_ad_read(int fd, char *buf, int nbytes, int chnl, int reuse);
extern int das_ad_cread(int fd, char *buf, int nbytes, int chnl, int reuse);
extern int das_ad_enqueue(int fd, int chnl, das_buf_t *buf, int nbytes, int mode, pid_t proxy);
extern das_buf_t * das_ad_dequeue(int fd, int chnl);
extern int das_ad_sync(int fd, int chnl, char *buf);
extern int das_ad_term(int fd, int chnl);
extern int das_ad_set_scan(int fd, char *scanlist, int n_entries);
extern int das_ad_get_range(int fd, int chnl, long *rmin, long *rmax);
extern int das_ad_set_range(int fd, int chnl, long rmin, long rmax);
extern int das_ad_set_trig(int fd, int chnl, int trig_src);
extern int das_ad_get_trig(int fd, int chnl, int *trig_src);
extern int das_ad_set_sample(int fd, int chnl, unsigned long ticks);
extern int das_ad_fqueue(int fd, int chnl);
extern int das_ad_rqueue(int fd, int chnl);

// Digital to Analog Conversion Routines

extern int das_da_sync(int fd, int chnl, char *buf);
extern int das_da_enqueue(int fd, int chnl, das_buf_t *buf, int nbytes, int mode, pid_t proxy);
extern int das_da_term(int fd, int chnl);
extern int das_da_cwait(int fd, int chnl, int reuse);
extern int das_da_wait(int fd, int chnl, int reuse);
extern int das_da_set_trig(int fd, int chnl, int trig_src);
extern int das_da_get_trig(int fd, int chnl, int *trig_src);
extern int das_da_get_range(int fd, int chnl, long *rmin, long *rmax);
extern int das_da_set_range(int fd, int chnl, long rmin, long rmax);
extern int das_da_set_sample(int fd, int chnl, unsigned long ticks);
extern int das_da_fqueue(int fd, int chnl);
extern int das_da_rqueue(int fd, int chnl);

extern int das_open(const char *path, int oflag, const char *cfg_file);
extern int das_close(int fd);
extern int das_get_info(int fd, das_info *info);

// Memory Routines 
extern das_buf_t * das_alloc(int fd, int nbufs, int nbytes, unsigned flags);
extern int das_free(int fd, das_buf_t *dptr);

#define _DAS_HDR_H_INCLUDED
#endif
