// Metrabyte DAS-20 adapter
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
// das_da.c module contains routines to:
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
//    perform a D/A conv via sw        - das_da_sync()
//    perform 'N' A/D conv via
//       interrupts or dma             - das_da_enqueue()
//    terminate conversions on a chnl  - das_da_term()
//    return the range/gain on a chnl  - das_da_get_range()
//    set the range/gain on a chnl     - das_da_set_range()
//    set the trigger source on a chnl - das_da_set_trig()
//    return the trig source on a chnl - das_da_get_trig()
//    set the conv rate on a chnl      - das_da_set_sample()
//    get the status on an conv        - das_da_cwait()
//    wait for a D/A request to finish - das_da_wait()
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

#include <i86.h>
#include <errno.h>
#include <stdio.h>
#include <conio.h>
#include <sys/kernel.h>
#include <met_20.h>
#include <das.h>
#include <das_extern.h>
#include <das_inline.h>
#include <das_struct.h>
#include <das_dma.h>
#include <das_tmr.h>


struct das_rng das_da_rng[NUM_DA_RANGES] = {
          {         0, 10000000 },        //    0V - 10V
          {  -5000000,  5000000 },        //   -5V - 5V
          {         0,        0 },        //    illegal position
          { -10000000, 10000000 } };      //  -10V - 10V


// Function das_da_sync performs a single digital to analog conversion
// Note the buffer must be as large as the D/A converter resolution.
// ie if the resolution is 12 bit, the buffer must be 2 bytes.
int das_da_sync(int fd, int chnl, char *buf)
{
   if(das_chk_init_fd(fd) != EOK)         // library initialized?
      return(-1);

   if(chnl > MAX_DA_CHNL)                 // valid channel?
   {
      errno = ECHRNG;
      return(-1);
   }

   if(buf == NULL)                        // valid pointer
   {
      errno = EINVAL;
      return(-1);
   }

   if(docb.da[0].r_hd)                    // currently in progress return busy
   {
      errno = EBUSY;
      return(-1);
   }

   outpw(docb.info.port+docb.da[chnl].addr, *(das_t *)buf);
   return(EOK);
}

// Function das_da_dma_sync polls the DMA controller for terminal count.
int das_da_dma_sync()
{
                                          // poll the dma controller
   while(read_dma(0, docb.info.dma_chnl) != 0xFFFF)
      ;
   outp(docb.info.port + INT_CTL_REG, INT_DISABLE);   // disable INT
   outp(docb.info.port + ADC_CTL, DMA_UNACTIVE);      // disable DMA
   outp(docb.info.port + DMA_CTL_REG, DMA_DISABLE);   // disable DMA
   das_ct_stop(_das_fd, 1);
   das_ct_stop(_das_fd, 2);

      //  unlink from request queue and link to finish queue
   das_link(das_unlink(0, &docb.da[0].r_hd, &docb.da[0].r_tl),
             &docb.da[0].f_hd, &docb.da[0].f_tl);
   docb.da[0].n_rent--;
   docb.da[0].n_fent++;
   docb.da[0].stat = COMPLETE;            // set status to complete
   Trigger(docb.da[0].f_tl->proxy);
   return(0);
}

// Function das_da_enqueue enqueues an D/A request.
int das_da_enqueue(int fd, int chnl, das_buf_t *buf, int nbytes, int mode, pid_t proxy)
{
   switch(mode & OPER_MASK)               // valid request operation?
   {     
      case DAS_INT:
      case DAS_DMA:
         return(das_da_request(fd, chnl, buf, nbytes, mode, proxy));

      case DAS_DMA_SYNC:
         if(das_da_request(fd, chnl, buf, nbytes, mode, proxy) !=EOK)
            return(-1);
         return(das_da_dma_sync());

      default:
         errno = EINVAL;
         return(-1);
   }
}

// Function das_da_request places a conversion request into the request
// queue associated with 'chnl'.
int das_da_request(int fd, int chnl, das_buf_t *buf, int nbytes, int mode, pid_t proxy)
{
   struct das_queue *ptr;

   if(das_chk_init_fd(fd) != EOK)               // libary initalized?
      return(-1);

   // let them pass in any channel, but put the request in the da[0] queue
   if(chnl > DA_CHNLS)
   {
      errno = ECHRNG;
      return(-1);
   }

   if(docb.ad[0].r_hd)                       // adapter doesn't support A/D
   {                                         // and D/A at the same time
      errno = EBUSY;
      return(-1);
   }

   // limit number of entries in the queues
   if((docb.da[0].n_rent + docb.da[0].n_fent) == MAX_QUEUE_ENT)
   {
      errno = EBUSY;
      return(-1);
   }

   // If there are request in the queue and mode is DMA_SYNC return busy
   if((mode & DAS_DMA_SYNC) && docb.da[0].r_hd)
   {
      errno = EBUSY;
      return(-1);
   }

   // If mode is INT and sample rate is greater than 10kHz return EINVAL
   if((mode & DAS_INT) && (docb.da[0].s_rate.div1 * docb.da[0].s_rate.div2)
          < INT_TICKS )
   {
      errno = EINVAL;
      return(-1);
   }

   if(buf->bufp == NULL)                     // valid buffer?
   {
      errno = EINVAL;
      return(-1);
   }

   // check for a valid number of conversions
   if(nbytes == 0 || nbytes > MAX_CONV)
   {
      errno = EINVAL;
      return(-1);
   }

   if(docb.da[0].stat == DAS_RECYCLE)     // no sense queueing any more if
   {                                      // a recycle is in the queue
      errno = EBUSY;
      return(-1);
   }

   if((ptr = das_create_entry()) == (struct das_queue *)NULL)
      return(-1);
   ptr->mode = mode & 0xFF;
   if(!(mode & DAS_DUAL_CHNL))               // if not in dual channel stuff
      ptr->mode |= (chnl << 8);              // the channel number in bit 8
   ptr->rtype = mode & 0xF00;
   ptr->proxy = proxy;
   ptr->bytes = ptr->cnt = nbytes;
   ptr->dma = buf;
   ptr->cp = (das_t *)buf->bufp;
   docb.da[0].n_rent++; 
   docb.da[0].stat = mode;
   if(das_link(ptr, &docb.da[0].r_hd, &docb.da[0].r_tl))
      das_da_cmd(0);
   return(EOK);
}

// Function das_da_term() stops D/A conversions on 'chnl'.
int das_da_term(int fd, int chnl)
{
   struct das_queue *r_hd, *f_hd, *r_tl, *f_tl;
   int psw;
   
   if(das_chk_init_fd(fd) != EOK)               // library initialized?
      return(-1);

   if(chnl > DA_CHNLS)                          // for adapters that support
   {                                            // dual issue, we use channel
      errno = ECHRNG;                           // 0 queues.
      return(-1);
   }

   das_ct_stop(_das_fd, 1);                     // stop the counters.
   das_ct_stop(_das_fd, 2);

   psw = disable();
   r_hd = docb.da[chnl].r_hd;
   f_hd = docb.da[chnl].f_hd;
   r_tl = docb.da[chnl].r_tl;
   f_tl = docb.da[chnl].f_tl;
   docb.da[chnl].r_hd = docb.da[chnl].r_tl = NULL;
   docb.da[chnl].f_hd = docb.da[chnl].f_tl = NULL;
   restore(psw);

   // terminate the current interrupt or dma conversions
   if(r_hd)
   {
      docb.da[0].stat = COMPLETE;
      outp(docb.info.port + INT_CTL_REG, INT_DISABLE);   // disable INT
      outp(docb.info.port + ADC_CTL, DMA_UNACTIVE);      // disable DMA
      outp(docb.info.port + DMA_CTL_REG, DMA_DISABLE);   // disable DMA
      if(r_hd->rtype == DAS_DMA)                         // stop DMA
         clear_dma(0, docb.info.dma_chnl);
   }
   docb.da[0].n_fent = 0;
   docb.da[0].n_rent = 0;

   while(r_hd)             // clean up the queues
      das_unlink(1, &r_hd, &r_tl);
   while(f_hd)
   {
      Creceive(f_hd->proxy, 0, 0);
      das_unlink(1, &f_hd, &f_tl);
   }
   return(EOK);
}

// Function das_da_cwait() returns the status of a D/A 'chnl.  Returns zero
// on success, -1 otherwise.
int das_da_cwait(int fd, int chnl, int reuse)
{
   int rtype;
   
   if(das_chk_init_fd(fd) != EOK)         // library initialized?
      return(-1);

   if(chnl > DA_CHNLS)
   {
      errno = ECHRNG;
      return(-1);
   }

   if(!docb.da[0].f_hd)                   // anything in the finish queue?
      return(0);

   if(reuse && docb.ad[0].r_hd)
   {
      das_unlink(1, &docb.da[0].f_hd, &docb.da[0].f_tl);
      docb.da[0].n_fent--;
      errno = EBUSY;
      return(-1);
   }

   if(reuse)                              // re-enqueue the request
   {

      rtype = docb.da[0].f_hd->rtype;
      docb.da[0].f_hd->cnt = docb.da[0].f_hd->bytes;
      docb.da[0].f_hd->cp = (das_t *)docb.da[0].f_hd->dma->bufp;
      docb.da[0].n_rent++; 
      docb.da[0].stat = docb.da[0].f_hd->mode;
         //  unlink from finish queue and link to ready queue
      if(das_link(das_unlink(0, &docb.da[0].f_hd, &docb.da[0].f_tl),
               &docb.da[0].r_hd, &docb.da[0].r_tl))
         das_da_cmd(0);
      if(rtype == DAS_DMA_SYNC)
         das_da_dma_sync();
   }
   else              // unlink request from finish queue and free its memory
      das_unlink(1, &docb.da[0].f_hd, &docb.da[0].f_tl);
   docb.da[0].n_fent--;
   return(1);                             // return the number of conversions
}

// Function das_da_wait() waits for a D/A request to finish.  A check
// is made to make sure they don't block if they haven't initiated any
// conversions.  The 'chnl' finish queue is checked for an entry.  If
// no entries exist in the queue, a loop is entered that checks for an entry,
// and calls Receive().  Once an entry is in the queue it is unlinked from
// it, the Creceive() is called to make sure proxies don't build up, and the
// number of conversions is returned.  This function only needs one proxy
// to manage 'N' D/A channel queues.  Returns zero on success, -1 otherwise.
int das_da_wait(int fd, int chnl, int reuse)
{
   int psw, rtype;
   pid_t proxy;
   
   if(das_chk_init_fd(fd) != EOK)
      return(-1);

   if(chnl > DA_CHNLS)
   {
      errno = ECHRNG;
      return(-1);
   }
   // Make sure they don't block if they haven't initiated any conversions
   psw = disable();
   if(docb.da[0].f_hd == NULL && docb.da[0].r_hd == NULL)
   {
      errno = EDEADLK;
      restore(psw);
      return(-1);
   }

   if(docb.da[0].stat == DAS_RECYCLE && docb.da[0].f_hd == NULL)
   {
      errno = EDEADLK;
      restore(psw);
      return(-1);
   }
   if(!docb.da[0].f_hd)
      proxy = docb.da[0].r_hd->proxy;
   else
      proxy = docb.da[0].f_hd->proxy;
   restore(psw);
   if(!proxy)                             // no proxy to block on
   {
      errno = EINVAL;
      return(-1);
   }

   while(!docb.da[0].f_hd)                   // wait for an entry in the queue
   {
      if(Receive(proxy, 0, 0 ) == -1)        // when an entry is put in the
         return(-1);                         // queue we will always get
   }                                         // get a proxy to unblock us.

   // Make sure a A/D request isn't in progress they want to reuse the buffer.
   if(reuse && docb.ad[0].r_hd)
   {
      das_unlink(1, &docb.da[0].f_hd, &docb.da[0].f_tl);
      docb.da[0].n_fent--;
      errno = EBUSY;
      return(-1);
   }
   
   if(reuse)
   {

      rtype = docb.da[0].f_hd->rtype;
      docb.da[0].f_hd->cnt = docb.da[0].f_hd->bytes;
      docb.da[0].f_hd->cp = (das_t *)docb.da[0].f_hd->dma->bufp;
      docb.da[0].n_rent++; 
      docb.da[0].stat = docb.da[0].f_hd->mode;

      //  unlink from finish queue and link to request queue
      if(das_link(das_unlink(0, &docb.da[0].f_hd, &docb.da[0].f_tl),
               &docb.da[0].r_hd, &docb.da[0].r_tl))
         das_da_cmd(0);
      if(rtype == DAS_DMA_SYNC)
         das_da_dma_sync();
   }
   else              // unlink request from finish queue and free its memory
      das_unlink(1, &docb.da[0].f_hd, &docb.da[0].f_tl);
   docb.da[0].n_fent--;
   return(EOK);                              // return the number of conversions
}

// Function das_da_get_range() returns the output range on D/A 'chnl'.
int das_da_get_range(int fd, int chnl, long *rmin, long *rmax)
{
   if(das_chk_init_fd(fd) != EOK)            // library initialized?
      return(-1);

   if(chnl > MAX_DA_CHNL)                    // valid channel?
   {
      errno = ECHRNG;
      return(-1);
   }

   docb.da[chnl].rng = (inp(docb.info.port + DMA_CTL_REG) >> chnl) & 0x05;
   docb.da[chnl].rng = (docb.da[chnl].rng >> 1) | (docb.da[chnl].rng & 0x01);

   *rmin = das_da_rng[docb.da[chnl].rng].min;
   *rmax = das_da_rng[docb.da[chnl].rng].max;

   return(EOK);
}

#pragma off (unreferenced)

// Function das_da_set_range() sets the range on D/A 'chnl'.
// Note this function is not supported by the DAS-20
int das_da_set_range(int fd, int chnl, long rmin, long rmax)
{
   errno = ENOTSUP;
   return(-1);
}

// Function das_da_get_trig() returns the trigger source on D/A 'chnl'.
int das_da_get_trig(int fd, int chnl, int *trig_src)
{
   if(das_chk_init_fd(fd) != EOK)            // library initialized?
      return(-1);

   *trig_src = docb.da[0].trig_src &0xFF;
   return(EOK);
}
#pragma on (unreferenced)

// Function das_da_set_trig() sets the trigger source on D/A 'chnl'.
int das_da_set_trig(int fd, int chnl, int trig_src)
{
   if(das_chk_init_fd(fd) != EOK)            // library initialized?
      return(-1);
   
   switch(trig_src)                          // valid trigger source?
   {
      case DAS_PC_TRIG:             // pacers are counters 1 and 2 gating on 1
         docb.cntrs.ctr[1-1].md.gtl = _NO_GATE;    // no gating
         break;
      case DAS_PC_TRIG | DAS_ENB_GATE | DAS_A_LOW:
         docb.cntrs.ctr[1-1].md.gtl = _LL_GATE_N;  // active low level gate N
         break;
      case DAS_PC_TRIG | DAS_ENB_GATE | DAS_A_HIGH:
         docb.cntrs.ctr[1-1].md.gtl = _HL_GATE_N;  // active high level gate N
         break;
      default:
         errno = EINVAL;
         return(-1);
   }
   docb.da[chnl].trig_src = (_HW_TRIG << 8) | trig_src;
   return(EOK);
}

// Function das_da_set_sample() sets the conversion rate and D/A 'chnl'.
int das_da_set_sample(int fd, int chnl, unsigned long ticks)
{
   if(das_chk_init_fd(fd) != EOK)               // library initialized?
      return(-1);

   if(ticks < MIN_TICKS || ticks > MAX_TICKS)   // valid sample rate?
   {
      errno = EINVAL;
      return(-1);
   }

   // convert ticks to two divisors that can be written to counters
   das_convert_rate(ticks, &docb.da[chnl].s_rate);
   docb.cntrs.ctr[1-1].ld = docb.da[chnl].s_rate.div2;
   docb.cntrs.ctr[2-1].ld = docb.da[chnl].s_rate.div1;

   if(docb.da[chnl].r_hd && (docb.da[chnl].trig_src & DAS_PC_TRIG))
   {
      das_ct_write(_das_fd, 1, &docb.cntrs.ctr[1-1]);
      das_ct_write(_das_fd, 2, &docb.cntrs.ctr[2-1]);
   }
   return(EOK);
}

// Function das_chk_init_fd() checks for a valid file descriptor and that the
// library has been initialized.
int das_chk_init_fd(int fd)
{
   if(!init_flag)
   {
      errno = EBADF;
      return(-1);
   }
   if(fd != _das_fd)
   {
      errno = EBADF;
      return(-1);
   }
   return(EOK);
}

// Function das_da_rqueue returns the number of requests in the ready queue
// for a 'chnl'.
int das_da_rqueue(int fd, int chnl)
{
   if(das_chk_init_fd(fd) != EOK)            // library initialized?
      return(-1);

   // valid channel?
   if(chnl > ((docb.info.chnl_cnfg == DAS_SINGLE_ENDED) ? MAX_SNGL_CHNL : MAX_DIF_CHNL))
   {
      errno = ECHRNG;
      return(-1);
   }
   return(docb.da[0].n_rent);
}

// Function das_da_fqueue returns the number of requests in the ready queue
// for a 'chnl'.
int das_da_fqueue(int fd, int chnl)
{
   if(das_chk_init_fd(fd) != EOK)            // library initialized?
      return(-1);

   // valid channel?
   if(chnl > ((docb.info.chnl_cnfg == DAS_SINGLE_ENDED) ? MAX_SNGL_CHNL : MAX_DIF_CHNL))
   {
      errno = ECHRNG;
      return(-1);
   }
   return(docb.da[0].n_fent);
}
