// das_dmaio.c module contains routines to:
//    program the DMA controller       - setup_dma()
//    start DMA on a channel           - kick_dma()
//    stop DMA on a channel            - clear_dma()
//    read the current count of a chnl - read_dma()

#include <sys/types.h>
#include <i86.h>
#include <das_inline.h>
#include <conio.h>
#include <das_dma.h>

/* 8237 mode bits */


#define DMA_M_OUT       (1 << 3)
#define DMA_M_IN        (1 << 2)

#define DMA_M_AUTO_INIT    (1 << 4)
#define DMA_M_NO_AUTO_INIT (0 << 4)

#define DMA_M_AUTO_DECR  (1 << 5)
#define DMA_M_AUTO_INCR  (0 << 5)

#define DMA_M_DEMAND    (0 << 6)
#define DMA_M_SINGLE    (1 << 6)
#define DMA_M_BLOCK     (2 << 6)
#define DMA_M_CASCADE   (3 << 6)

/* MASK BITS */
#define DMA_MASK_SET    (1 << 2)
#define DMA_MASK_CLEAR  (0 << 2)


/*    PS/2 dma function registers */
#define PS2E_REG  0x18
#define PS2E_EXE  0x1a

/*    PS/2 extended mode register */
#define PS2_EMODE_READ  0x00
#define PS2_EMODE_WRITE 0x08

#define PS2_EMODE_BYTE  0x00
#define PS2_EMODE_WORD  0x40

#define PS2_EMODE_DOXFER 0x04
#define PS2_EMODE_DOVERR 0x00

#define PS2_EMODE_DO_IO  0x01
#define PS2_EMODE_DO_DRQ 0x00

#define PS2E_UNMASK_CHANNEL (0x0A << 4)
#define PS2E_MASK_CHANNEL   (0x09 << 4)
#define PS2E_SETMODE        (0x07 << 4)
#define PS2E_READSTATUS     (0x06 << 4)
#define PS2E_READCOUNT      (0x05 << 4)
#define PS2E_WRITECOUNT     (0x04 << 4)
#define PS2E_READADDR       (0x03 << 4)
#define PS2E_WRITEADDR      (0x02 << 4)


struct    regs_8237 {
        int  maddr;   /* memory address register */
        int  tcount;  /* terminal count register */
        int  pagereg; /* dma page select register */
        int  stat;    /* status register */
        int  mask;    /* mask register */
        int  mode;    /* mode register */
        /* 
           these three are psuedo-registers, they don't care what you write
           to them, it's just that you wrote them
        */
        int  sync_cmd;   /* Synchronize HI/LOW register access */
        int  reset_cmd;  /* Reset the 8237 -- bad news, don't know how to init */
        int  enable_cmd; /* enable all dma channels */
        int  set_mask;   /* write mask register */
    };

/*    at dma registers    */
static const struct regs_8237 regtable[] = {
        /*   addr  count page  stat/cmd mask  mode  clr   rst   enable wmask */
        {    0x00, 0x01, 0x87, 0x08,    0x0a, 0x0b, 0x0c, 0x0d, 0x0e, 0x0f },
        {    0x02, 0x03, 0x83, 0x08,    0x0a, 0x0b, 0x0c, 0x0d, 0x0e, 0x0f },
        {    0x04, 0x05, 0x81, 0x08,    0x0a, 0x0b, 0x0c, 0x0d, 0x0e, 0x0f },
        {    0x06, 0x07, 0x82, 0x08,    0x0a, 0x0b, 0x0c, 0x0d, 0x0e, 0x0f },


        {    0xc0, 0xc2, 0x8f, 0xd0,    0xd4, 0xd6, 0xd8, 0xda, 0xdc, 0xde },
        {    0xc4, 0xc6, 0x8b, 0xd0,    0xd4, 0xd6, 0xd8, 0xda, 0xdc, 0xde },
        {    0xc8, 0xca, 0x89, 0xd0,    0xd4, 0xd6, 0xd8, 0xda, 0xdc, 0xde },
        {    0xcc, 0xce, 0x8a, 0xd0,    0xd4, 0xd6, 0xd8, 0xda, 0xdc, 0xde }
    };
                                                   

setup_dma(unsigned dflags, dma_t *iobufp, unsigned iosize, int channel)
{
int    psw;
struct    regs_8237    *dp = regtable + channel;
union {
    paddr_t    addr;
    unsigned char bytes[3];
} physaddr;


    if (channel > 7)
        return -1;        /*    only 0-7 are supported */

    physaddr.addr = iobufp->phys_addr;

    psw = disable();

	clear_dma(dflags, channel);
    if (dflags & DMA_PS2) {
/* Set channel mask bit        */

        outp(PS2E_REG,  PS2E_MASK_CHANNEL | channel);            
        outp(PS2E_REG,  PS2E_WRITEADDR    | channel);            
/* Setup to write mem addr    */
        outp(PS2E_EXE, physaddr.bytes[0]);
        outp(PS2E_EXE, physaddr.bytes[1]);
        outp(PS2E_EXE, physaddr.bytes[2]);
/* Setup to write count        */
        outp(PS2E_REG,  PS2E_WRITECOUNT | channel);
        if ((dflags & DMA_WORD) == 0) {
            iosize /= sizeof(int)/sizeof(char);
        }
        --iosize;
        outp(PS2E_EXE, iosize);
        outp(PS2E_EXE, (iosize >> 8) & 0xff);
/* Setup to write ext mode    */
        outp(PS2E_REG,  PS2E_SETMODE | channel);
        outp(PS2E_EXE, 
            ((dflags & DMA_OUTPUT) ? PS2_EMODE_WRITE : PS2_EMODE_READ) |
            ((dflags & DMA_WORD) ? PS2_EMODE_WORD : PS2_EMODE_BYTE) |
            PS2_EMODE_DOXFER);
    } else {
        if (dflags & DMA_WORD)
            iosize *= sizeof(int)/sizeof(char);
        iosize--;
        outp(dp->sync_cmd,0);
        /* set dma mode, single, auto_incr, no_init, out-or-in */
        outp(dp->mode, DMA_M_SINGLE | DMA_M_AUTO_INCR |
             ((dflags & DMA_AUTO) ? DMA_M_AUTO_INIT : DMA_M_NO_AUTO_INIT) |
             ((dflags & DMA_OUTPUT) ? DMA_M_OUT : DMA_M_IN) |
             (channel & 0x03));
   
        outp(dp->maddr,   physaddr.bytes[0]);
        outp(dp->maddr,   physaddr.bytes[1]);
        outp(dp->pagereg, physaddr.bytes[2]);
        outp(dp->tcount,  iosize);
        outp(dp->tcount,  (iosize >> 8) & 0xff);
    }
    restore(psw);
    return (dflags & DMA_WAIT) ? 0 : kick_dma(dflags,channel);
}

kick_dma(unsigned dflags, int channel)
{
    if (dflags & DMA_PS2) {
        outp(PS2E_REG,  PS2E_UNMASK_CHANNEL | channel);
    } else {
        outp((regtable+channel)->mask, DMA_MASK_CLEAR | (channel & 3));
    }

    return 0;
}

clear_dma(unsigned dflags, int channel)
{
    if (dflags & DMA_PS2) {
        outp(PS2E_REG, PS2E_MASK_CHANNEL | channel);
    } else {
        outp(regtable[channel].mask, DMA_MASK_SET | (channel & 3));
    }
    return 0;
}

read_dma(unsigned dflags, int channel)
{
int psw = disable();
int rcnt;
     if (dflags & DMA_PS2) {
        outp(PS2E_REG, PS2E_READCOUNT | channel);
        rcnt = inp(PS2E_EXE);
        rcnt += inp(PS2E_EXE) << 8;
     } else {
        outp((regtable+channel)->sync_cmd,0);
        rcnt = inp((regtable+channel)->tcount);
        rcnt += inp((regtable+channel)->tcount) << 8;
     }
     restore(psw);
     return rcnt;
}
