// Metrabyte DAS-20 adapter
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
// das_tmr.c module contains routines to:
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
//    set a cntr to a specific mode    - das_ct_mode()
//    get the status of a cntr         - das_ct_status()
//    read the current count of a cntr - das_ct_read()
//    write the count to a cntr        - das_ct_write()           
//    stop the counting on a cntr      - das_ct_stop()
//    convert Hz to cntr values        - das_convert_rate()
//    set the data ptr to a register   - das_point_to()
//    set the master mode register     - das_set_master()
//    perform a command on a counter   - das_ctr_cmd()
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
// If the memory model isn't small, compile with the -zu option

#include <met_20.h>
#include <das_struct.h>
#include <das_extern.h>
#include <das_tmr.h>
#include <stdio.h>
#include <errno.h>
#include <conio.h>
#include <das_inline.h>

#pragma off (check_stack)
#pragma off (unreferenced)

// This function is called to initialize 'n_ctr' to a mode specified by
// the paramaters pointed to by 'cptr'.
int das_ct_mode(int fd, int n_ctr, cntr_mdr_t *cptr, int clk_src)
{
   das_point_to(n_ctr, _MODE);
   outp(docb.info.port + CTR_DATA, cptr->all);
   outp(docb.info.port + CTR_DATA, cptr->all >> 8);
   return(EOK);
}   
 
// The status of the counter specified by 'n_ctr' is return in 'status'.
// status register format is:    SR7         0 or 1 undefined
//                               SR6         0 or 1 undefined
//                               SR5         output level of counter 5
//                               SR4         output level of counter 4
//                               SR3         output level of counter 3
//                               SR2         output level of counter 2
//                               SR1         output level of counter 1
//                               SR0         byte pointer
// If counters 1 and 2 comparators are enable the output level is from the
// comparators not TC.
int das_ct_status(int fd, int n_ctr, int *status)
{
   static int ctr_bits[5] = { _CTR1, _CTR2, _CTR3, _CTR4, _CTR5 };
   if (n_ctr > MAX_CTR || n_ctr == 0)
   {
      errno = EINVAL;
      return(-1);
   }
    
   *status = inp(docb.info.port + CTR_CTL) & (ctr_bits[n_ctr-1]<<1);
   *status >>= ctr_bits[n_ctr-1];
   return(EOK);
}

// Returns the current count of the counter specified by 'n_ctr'. 
int das_ct_read(int fd, int n_ctr, int *cnt)
{
   if(n_ctr > MAX_CTR || n_ctr == 0)
   {
      errno = EINVAL;
      return(-1);
   }
   das_ctr_cmd(n_ctr, _S_CTR);                  // save count in hold register
   das_point_to(n_ctr, _HOLD);                        // set pointer to hold reg
   *cnt = inp(docb.info.port + CTR_DATA) | (inp(docb.info.port + CTR_DATA) << 8);
   return(EOK);
}

// Writes the count specified by '*cptr' to the counter specified by 'n_ctr'.
int das_ct_write(int fd, int n_ctr, cntr_t *cptr)
{
   if (n_ctr > MAX_CTR || n_ctr == 0)
   {
      errno = EINVAL;
      return(-1);
   }
   das_point_to(n_ctr, _LOAD);                           // set pointer to 'n_ctr'
   outp(docb.info.port + CTR_DATA, cptr->ld);      // set the load register
   outp(docb.info.port + CTR_DATA, cptr->ld >> 8);
   if(cptr->hd)                                    // set the hold register
   {
      outp(docb.info.port + CTR_DATA, cptr->hd);
      outp(docb.info.port + CTR_DATA, cptr->hd >> 8);
   }
   
   das_ctr_cmd(n_ctr, _LD_CTR);                    // load the value
   das_ctr_cmd(n_ctr, _ARM_CTR);                   // arm the counter
   return(EOK);
}

// Stops counting operation on counter specified by 'n_ctr'.
int das_ct_stop(int fd, int n_ctr)
{
   if(n_ctr > MAX_CTR || n_ctr == 0)
   {
      errno = EINVAL;
      return(-1);
   }
   das_ctr_cmd(n_ctr, _DARM_CTR);
   return(EOK);
}

// Function das_point_to() sets the data pointer of the AMD9513 to a register.
void das_point_to(int chnl, int reg)
{
   cmd_t _cmd;

   _cmd.dcmd.grp = chnl;
   _cmd.dcmd.elm = reg;
   _cmd.dcmd.cmd = _LD_DPTR;
   outp(docb.info.port + CTR_CTL, _cmd.cmd);
}

// Function das_set_master() sets the master mode register on the AMD9513.
int das_set_master(int fd, cntr_mmdr_t *md)
{
   outp(docb.info.port + CTR_CTL, _MRESET);     // reset the STC
   das_point_to(_CTL_GRP, _MASTER);
   outp(docb.info.port + CTR_DATA, md->all);
   outp(docb.info.port + CTR_DATA, (md->all >> 8));
   return(EOK);
}

// Function das_ctr_cmd() performs a command on 'ctr' eg disarm counter.
int das_ctr_cmd(int ctr, int cmd)
{
   static int ctr_bits[5] = { _CTR1, _CTR2, _CTR3, _CTR4, _CTR5 };
   cmd_t _cmd;
   
   _cmd.ccmd.ctr = ctr_bits[ctr-1];
   _cmd.ccmd.cmd = cmd;
   outp(docb.info.port + CTR_CTL, _cmd.cmd);
   return(EOK);
}
#pragma on (unreferenced)
#pragma on (check_stack)
