//*****************************************************************************
// Module das_ad_msg.c contains the messaging functions for the 'das' resource
// manager driver.  These functions send all A/D conversion operation messages 
// to the resource manager though a file descriptor.
//*****************************************************************************
// The follwing is the list of functions in this module.
//
//    das_ad_sync
//    das_ad_read
//    das_ad_cread
//    das_ad_term
//    das_ad_set_scan
//    das_ad_get_range
//    das_ad_set_range
//    das_ad_get_trig
//    das_ad_set_trig
//    das_ad_set_sample
//*****************************************************************************
#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include <sys/kernel.h>
#include <sys/sendmx.h>
#include <das.h>
#include "das_struct_msg.h"

extern union das_msg    io_msg;
extern das_info das_query_info;

//*****************************************************************************
// das_ad_sync -  Send message to preform a polled A/D conversion (blocking).
//*****************************************************************************
int das_ad_sync(int fd, int chnl, char *data)
{
      struct _mxfer_entry  mx[3];
                     
                     // Fill in message structure.
   io_msg.das_ad_sync.type    = _IO_DAS;
   io_msg.das_ad_sync.request = _DAS_AD_SYNC; 
   io_msg.das_ad_sync.fd      = fd; 
   io_msg.das_ad_sync.n_bytes = das_query_info.ad_res;
   io_msg.das_ad_sync.chnl    = chnl;
   
                     // Send message header.
   _setmx(&mx[0], &io_msg, sizeof(io_msg.das_ad_sync));
                     // Reply message header.
   _setmx(&mx[1], &io_msg, sizeof(io_msg.das_ad_sync_reply) -
         sizeof(io_msg.das_ad_sync_reply.data));   
                     // Reply data.
   _setmx(&mx[2], data, das_query_info.ad_res);            

   if(Sendfdmx(fd, 1, 2, &mx, &mx[1]) == -1)
      return(-1);
   if(io_msg.status != EOK)
   {
      errno = io_msg.status;
      return(-1);
   }
   return(0);
}

//*****************************************************************************
// das_ad_enqueue -  
//*****************************************************************************
#pragma off (unreferenced)
int das_ad_enqueue(int fd, int chnl, das_buf_t *ptr, int n_bytes, int mode,
                     pid_t proxy)
{
                     // Fill in message structure.
   io_msg.das_ad_enque.type     = _IO_DAS;
   io_msg.das_ad_enque.request  = _DAS_AD_ENQUE; 
   io_msg.das_ad_enque.fd       = fd; 
   io_msg.das_ad_enque.mode     = mode; 
   io_msg.das_ad_enque.n_bytes  = n_bytes;
   io_msg.das_ad_enque.chnl     = chnl;
   io_msg.das_ad_enque.proxy    = proxy;

   if(Sendfd(fd, &io_msg, &io_msg, sizeof(io_msg.das_ad_enque), 
               sizeof(io_msg.das_ad_enque_reply)) == -1)
      return(-1);
   if(io_msg.status != EOK)
   {
      errno = io_msg.status;
      return(-1);
   }
   return(0);
}

//*****************************************************************************
// das_ad_dequeue -  
//*****************************************************************************
das_buf_t * das_ad_dequeue(int fd, int chnl)
{
   errno = ENOREMOTE;
   return(NULL);
}
#pragma on (unreferenced)

//*****************************************************************************
// das_ad_read -  Block on a A/D operation until it is done, then read the data
//             from the resource manager.
//*****************************************************************************
int das_ad_read(int fd, char * buf, int n_bytes, int chnl, int reuse)
{
   struct _mxfer_entry  mx[3];

                     // Fill in message structure.
   io_msg.das_ad_read.type       = _IO_DAS;
   io_msg.das_ad_read.request    = _DAS_AD_READ; 
   io_msg.das_ad_read.fd         = fd;
   io_msg.das_ad_read.chnl       = chnl;
   io_msg.das_ad_read.n_bytes    = n_bytes;
   io_msg.das_ad_read.mode       = reuse;

                     // Send message header
   _setmx(&mx[0], &io_msg, sizeof(io_msg.das_ad_read));

                     // Reply message header
   _setmx(&mx[1], &io_msg, sizeof(io_msg.das_ad_read_reply) -
            sizeof(io_msg.das_ad_read_reply.data));          
                     // Reply data                             
   _setmx(&mx[2], buf, n_bytes);
            
   if(Sendfdmx(fd, 1, 2, &mx, &mx[1]) == -1)
      return(-1);
   if(io_msg.status != EOK)
   {
      errno = io_msg.status;
      return(-1);
   }
   return(io_msg.das_ad_read_reply.n_bytes);
}

//*****************************************************************************
// das_ad_cread - Perform a Crecieve on the A/D finish proxy to check the status
//             of a A/D operation without blocking.  If the operation is 
//             finished then read the data from the resource manager.
//*****************************************************************************
int das_ad_cread(int fd, char *buf, int n_bytes, int chnl, int reuse)
{
   struct _mxfer_entry  mx[3];

                     // Fill in message structure.
   io_msg.das_ad_read.type       = _IO_DAS;
   io_msg.das_ad_read.request    = _DAS_AD_CREAD; 
   io_msg.das_ad_read.fd         = fd;
   io_msg.das_ad_read.chnl       = chnl;
   io_msg.das_ad_read.n_bytes    = n_bytes;
   io_msg.das_ad_read.mode       = reuse;

                     // Send message header
   _setmx(&mx[0], &io_msg, sizeof(io_msg.das_ad_read));

                     // Reply messager header
   _setmx(&mx[1], &io_msg, sizeof(io_msg.das_ad_read_reply) -
            sizeof(io_msg.das_ad_read_reply.data));          
                     // Reply data.
   _setmx(&mx[2], buf, n_bytes);
            
   if(Sendfdmx(fd, 1, 2, &mx, &mx[1]) == -1)
      return(-1);
   if(io_msg.status != EOK)
   {
      errno = io_msg.status;
      return(-1);
   }
   return(io_msg.das_ad_read_reply.n_bytes);
}

//*****************************************************************************
// das_ad_rqueue -   
//*****************************************************************************
int das_ad_rqueue(int fd, int chnl)
{
                     // Fill in message structure.
   io_msg.das_ad_get_queue.type    = _IO_DAS;
   io_msg.das_ad_get_queue.request = _DAS_AD_GET_RDY_QUE; 
   io_msg.das_ad_get_queue.fd      = fd;
   io_msg.das_ad_get_queue.chnl    = chnl;

   if(Sendfd(fd, &io_msg, &io_msg, sizeof(io_msg.das_ad_get_queue), 
               sizeof(io_msg.das_ad_get_queue_reply)) == -1)
      return(-1);
   if(io_msg.status != EOK)
   {
      errno = io_msg.status;
      return(-1);
   }
   return(io_msg.das_ad_get_queue_reply.n_ent);
}

//*****************************************************************************
// das_ad_fqueue -   
//*****************************************************************************
int das_ad_fqueue(int fd, int chnl)
{
                     // Fill in message structure.
   io_msg.das_ad_get_queue.type    = _IO_DAS;
   io_msg.das_ad_get_queue.request = _DAS_AD_GET_FIN_QUE; 
   io_msg.das_ad_get_queue.fd      = fd;
   io_msg.das_ad_get_queue.chnl    = chnl;

   if(Sendfd(fd, &io_msg, &io_msg, sizeof(io_msg.das_ad_get_queue),
               sizeof(io_msg.das_ad_get_queue_reply)) == -1)
      return(-1);
   if(io_msg.status != EOK)
   {
      errno = io_msg.status;
      return(-1);
   }
   return(io_msg.das_ad_get_queue_reply.n_ent);
}

//*****************************************************************************
// das_ad_term -  Send message to terminate any A/D conversions currently in
//                progress, finished or queued.
//*****************************************************************************
int das_ad_term(int fd, int chnl)
{
                     // Fill in message structure.
   io_msg.das_ad_term.type    = _IO_DAS;
   io_msg.das_ad_term.request = _DAS_AD_TERM; 
   io_msg.das_ad_term.fd      = fd;
   io_msg.das_ad_term.chnl    = chnl;

   if(Sendfd(fd, &io_msg, &io_msg, sizeof(io_msg.das_ad_term), 
               sizeof(io_msg.das_ad_term_reply)) == -1)
      return(-1);
   if(io_msg.status != EOK)
   {
      errno = io_msg.status;
      return(-1);
   }
   return(0);
}

//*****************************************************************************
// das_ad_set_scan - Send message to set a block scan entry for A/D operations.
//*****************************************************************************
int das_ad_set_scan(int fd, char *scanlist, int n_entries)
{
   struct _mxfer_entry mx[3];
   
                     // Fill in message structure.
   io_msg.das_ad_scan.type       = _IO_DAS;
   io_msg.das_ad_scan.request    = _DAS_AD_SET_SCAN; 
   io_msg.das_ad_scan.fd         = fd;
   io_msg.das_ad_scan.n_entry    = n_entries;
   
                     // Send message header.
   _setmx(&mx[0], &io_msg, sizeof(io_msg.das_ad_scan) - 
            sizeof(io_msg.das_ad_scan.data));
                     // Send message data.
   _setmx(&mx[1], scanlist, n_entries);
   
                     // Reply message header.
   _setmx(&mx[2], &io_msg, sizeof(io_msg.das_ad_scan_reply));

   if(Sendfdmx(fd, 2, 1, &mx, &mx[2]) == -1)
      return(-1);
   if(io_msg.status != EOK)
   {
      errno = io_msg.status;
      return(-1);
   }
   return(0);
}

//*****************************************************************************
// das_ad_get_range -   Send message to get the current range (gain) setting 
//                      on a specific A/D channel.
//*****************************************************************************
int das_ad_get_range(int fd, int chnl, long *range_min, long *range_max)
{
                     // Fill in message structure.
   io_msg.das_ad_get_range.type    = _IO_DAS;
   io_msg.das_ad_get_range.request = _DAS_AD_GET_RANGE; 
   io_msg.das_ad_get_range.fd      = fd;
   io_msg.das_ad_get_range.chnl    = chnl;

   if(Sendfd(fd, &io_msg, &io_msg, sizeof(io_msg.das_ad_get_range),
               sizeof(io_msg.das_ad_get_range_reply)) == -1)
      return(-1);
   if(io_msg.status != EOK)
   {
      errno = io_msg.status;
      return(-1);
   }
   *range_min = io_msg.das_ad_get_range_reply.range_min;
   *range_max = io_msg.das_ad_get_range_reply.range_max;
   return(0);
}

//*****************************************************************************
// das_ad_set_range -   Send message to set a new range (gain) setting on a 
//                      specific A/D channel.
//*****************************************************************************
int das_ad_set_range(int fd, int chnl, long range_min, long range_max)
{
                     // Fill in message structure.
   io_msg.das_ad_set_range.type        = _IO_DAS;
   io_msg.das_ad_set_range.request     = _DAS_AD_SET_RANGE; 
   io_msg.das_ad_set_range.fd          = fd;
   io_msg.das_ad_set_range.chnl        = chnl;
   io_msg.das_ad_set_range.range_min   = range_min;
   io_msg.das_ad_set_range.range_max   = range_max;

   if(Sendfd(fd, &io_msg, &io_msg, sizeof(io_msg.das_ad_set_range), 
               sizeof(io_msg.das_ad_set_range_reply)) == -1)
      return(-1);
   if(io_msg.status != EOK)
   {
      errno = io_msg.status;
      return(-1);
   }
   return(0);
}

//*****************************************************************************
// das_ad_set_trig - Send message to set a new trigger source for A/D 
//                   operations.
//*****************************************************************************
int das_ad_set_trig(int fd, int chnl, int trig_src)
{
                     // Fill in message structure.
   io_msg.das_ad_set_trig.type       = _IO_DAS;
   io_msg.das_ad_set_trig.request    = _DAS_AD_SET_TRIG; 
   io_msg.das_ad_set_trig.fd         = fd;
   io_msg.das_ad_set_trig.trig_src   = trig_src;
   io_msg.das_ad_set_trig.chnl       = chnl;

   if(Sendfd(fd, &io_msg, &io_msg, sizeof(io_msg.das_ad_set_trig), 
               sizeof(io_msg.das_ad_set_trig_reply)) == -1)
      return(-1);
   if(io_msg.status != EOK)
   {
      errno = io_msg.status;
      return(-1);
   }
   return(0);
}

//*****************************************************************************
// das_ad_get_trig - Send message to get the current trigger source for A/D
//                   operations.
//*****************************************************************************
int das_ad_get_trig(int fd, int chnl, int *trig_src)
{
                     // Fill in message structure.
   io_msg.das_ad_get_trig.type       = _IO_DAS;
   io_msg.das_ad_get_trig.request    = _DAS_AD_GET_TRIG; 
   io_msg.das_ad_get_trig.fd         = fd;
   io_msg.das_ad_get_trig.chnl       = chnl;

   if(Sendfd(fd, &io_msg, &io_msg, sizeof(io_msg.das_ad_get_trig), 
               sizeof(io_msg.das_ad_get_trig_reply)) == -1)
      return(-1);
   if(io_msg.status != EOK)
   {
      errno = io_msg.status;
      return(-1);
   }
   *trig_src = io_msg.das_ad_get_trig_reply.trig_src;
   return(0);
}

//*****************************************************************************
// das_ad_set_sample -  Send message to set the sampling rate for A/D 
//                      operations.
//*****************************************************************************
int das_ad_set_sample(int fd, int chnl, long ticks)
{
                     // Fill in messsage structure.
   io_msg.das_ad_sam.type    = _IO_DAS;
   io_msg.das_ad_sam.request = _DAS_AD_SET_SMP; 
   io_msg.das_ad_sam.fd      = fd;
   io_msg.das_ad_sam.ticks   = ticks;
   io_msg.das_ad_sam.chnl    = chnl;

   if(Sendfd(fd, &io_msg, &io_msg, sizeof(io_msg.das_ad_sam), 
               sizeof(io_msg.das_ad_sam_reply)) == -1)
      return(-1);
   if(io_msg.status != EOK)
   {
      errno = io_msg.status;
      return(-1);
   }
   return(0);
}
