//*****************************************************************************
// Module das_da_msg.c contains the messaging functions for the 'das' resource
// manager driver.  These functions send all D/A conversion operation messages 
// to the resource manager though a file descriptor.
//*****************************************************************************
// The follwing is the list of functions in this module.
//
//    das_da_sync
//    das_da_stat
//    das_da_wait
//    das_da_term
//    das_da_get_range
//    das_da_set_range
//    das_da_get_trig
//    das_da_set_trig
//    das_da_set_sample
//*****************************************************************************
#include <errno.h>
#include <sys/kernel.h>
#include <sys/sendmx.h>
#include <das.h>
#include "das_struct_msg.h"

extern union das_msg    io_msg;
extern das_info das_query_info;

//*****************************************************************************
// das_da_sync -  Send a message to perform a polled D/A conversion. 
//*****************************************************************************
int das_da_sync(int fd, int chnl, char *data)
{
   struct _mxfer_entry  mx[2];
                     
                     // Fill in message structure.
   io_msg.das_da_sync.type    = _IO_DAS;
   io_msg.das_da_sync.request = _DAS_DA_SYNC; 
   io_msg.das_da_sync.fd      = fd; 
   io_msg.das_da_sync.n_bytes = das_query_info.da_res;
   io_msg.das_da_sync.chnl    = chnl;

                      // Send message header.
   _setmx(&mx[0], &io_msg, sizeof(io_msg.das_da_sync) - 
            sizeof(io_msg.das_da_sync.data));

                     // Send message data.
   _setmx(&mx[1], data, das_query_info.da_res);

   if(Sendfdmx(fd, 2, 1, &mx, &mx) == -1)
      return(-1);
   if(io_msg.status != EOK)
   {
      errno = io_msg.status;
      return(-1);
   }
   return(0);
}

//*****************************************************************************
// das_da_enqueue -
//*****************************************************************************
int das_da_enqueue(int fd, int chnl, das_buf_t *ptr, int n_bytes, int mode, 
                     pid_t proxy)
{
   struct _mxfer_entry  mx[3];
                     
                     // Fill in the message structure.
   io_msg.das_da_enque.type = _IO_DAS;
   io_msg.das_da_enque.request = _DAS_DA_ENQUE; 
   io_msg.das_da_enque.fd = fd; 
   io_msg.das_da_enque.mode = mode; 
   io_msg.das_da_enque.chnl = chnl; 
   io_msg.das_da_enque.n_bytes = n_bytes;
   io_msg.das_da_enque.proxy = proxy;

                     // Send message header.
   _setmx(&mx[0], &io_msg, sizeof(io_msg.das_da_enque) - 
            sizeof(io_msg.das_da_enque.data));
                     // Send message data.
   _setmx(&mx[1], ptr->bufp, n_bytes);
      
                     // Reply message header.
   _setmx(&mx[2], &io_msg, sizeof(io_msg.das_da_enque_reply));

   if(Sendfdmx(fd, 2, 1, &mx, &mx[2]) == -1)
      return(-1);
   if(io_msg.status != EOK)
   {
      errno = io_msg.status;
      return(-1);
   }
   return(0);
}

//*****************************************************************************
// das_da_term -  Send message to terminate any D/A conversions currently in
//                progress or queued.
//*****************************************************************************
int das_da_term(int fd, int chnl)
{
                     // Fill in message structure.
   io_msg.das_da_term.type    = _IO_DAS;
   io_msg.das_da_term.request = _DAS_DA_TERM; 
   io_msg.das_da_term.fd      = fd;
   io_msg.das_da_term.chnl    = chnl;

   if(Sendfd(fd, &io_msg, &io_msg, sizeof(io_msg.das_da_term), 
               sizeof(io_msg.das_da_term_reply)) == -1)
      return(-1);
   if(io_msg.status != EOK)
   {
      errno = io_msg.status;
      return(-1);
   }
   return(0);
}

//*****************************************************************************
// das_da_cwait - Check the status of a D/A conversion.  Returns -1 if not done
//                of 0 if done.
//*****************************************************************************
int das_da_cwait(int fd, int chnl, int reuse)
{
                     // Fill in message structure.
   io_msg.das_da_wait.type    = _IO_DAS;
   io_msg.das_da_wait.request = _DAS_DA_CWAIT; 
   io_msg.das_da_wait.fd      = fd;
   io_msg.das_da_wait.chnl    = chnl;
   io_msg.das_da_wait.reuse   = reuse;

   if(Sendfd(fd, &io_msg, &io_msg, sizeof(io_msg.das_da_wait), 
               sizeof(io_msg.das_da_wait_reply)) == -1)
      return(-1);
   if(io_msg.status != EOK)
   {
      errno = io_msg.status;
      return(-1);
   }
   return(io_msg.das_da_wait_reply.n_bytes);
}

//*****************************************************************************
// das_da_wait - Block on a D/A conversion until it finishes.
//*****************************************************************************
int das_da_wait(int fd, int chnl, int reuse)
{
                     // Fill in message structure.
   io_msg.das_da_wait.type    = _IO_DAS;
   io_msg.das_da_wait.request = _DAS_DA_WAIT; 
   io_msg.das_da_wait.fd      = fd;
   io_msg.das_da_wait.chnl    = chnl;
   io_msg.das_da_wait.reuse   = reuse;

   if(Sendfd(fd, &io_msg, &io_msg, sizeof(io_msg.das_da_wait), 
               sizeof(io_msg.das_da_wait_reply)) == -1)
      return(-1);
   if(io_msg.status != EOK)
   {
      errno = io_msg.status;
      return(-1);
   }
   return(io_msg.das_da_wait_reply.n_bytes);
}

//*****************************************************************************
// das_da_rqueue -   
//*****************************************************************************
int das_da_rqueue(int fd, int chnl)
{
                     // Fill in message structure.
   io_msg.das_da_get_queue.type    = _IO_DAS;
   io_msg.das_da_get_queue.request = _DAS_DA_GET_RDY_QUE; 
   io_msg.das_da_get_queue.fd      = fd;
   io_msg.das_da_get_queue.chnl    = chnl;

   if(Sendfd(fd, &io_msg, &io_msg, sizeof(io_msg.das_da_get_queue), 
               sizeof(io_msg.das_da_get_queue_reply)) == -1)
      return(-1);
   if(io_msg.status != EOK)
   {
      errno = io_msg.status;
      return(-1);
   }
   return(io_msg.das_da_get_queue_reply.n_ent);
}

//*****************************************************************************
// das_da_fqueue -   
//*****************************************************************************
int das_da_fqueue(int fd, int chnl)
{
                     // Fill in message structure.
   io_msg.das_da_get_queue.type    = _IO_DAS;
   io_msg.das_da_get_queue.request = _DAS_DA_GET_FIN_QUE; 
   io_msg.das_da_get_queue.fd      = fd;
   io_msg.das_da_get_queue.chnl    = chnl;

   if(Sendfd(fd, &io_msg, &io_msg, sizeof(io_msg.das_da_get_queue), 
               sizeof(io_msg.das_da_get_queue_reply)) == -1)
      return(-1);
   if(io_msg.status != EOK)
   {
      errno = io_msg.status;
      return(-1);
   }
   return(io_msg.das_da_get_queue_reply.n_ent);
}

//*****************************************************************************
// das_da_get_range -   Send message to get the current range (gain) setting 
//                      on a specific D/A channel.
//*****************************************************************************
int das_da_get_range(int fd, int chnl, long *range_min, long *range_max)
{
                     // Fill in message structure.
   io_msg.das_da_get_range.type    = _IO_DAS;
   io_msg.das_da_get_range.request = _DAS_DA_GET_RANGE; 
   io_msg.das_da_get_range.fd      = fd;
   io_msg.das_da_get_range.chnl    = chnl;

   if(Sendfd(fd, &io_msg, &io_msg, sizeof(io_msg.das_da_get_range), 
               sizeof(io_msg.das_da_get_range_reply)) == -1)
      return(-1);
   if(io_msg.status != EOK)
   {
      errno = io_msg.status;
      return(-1);
   }
   *range_min = io_msg.das_da_get_range_reply.range_min;
   *range_max = io_msg.das_da_get_range_reply.range_max;
   return(0);
}

//*****************************************************************************
// das_da_set_range -   Send message to set a new range (gain) setting on a 
//                      specific D/A channel.
//*****************************************************************************
int das_da_set_range(int fd, int chnl, long range_min, long range_max)
{
                     // Fill in message structue.
   io_msg.das_da_set_range.type        = _IO_DAS;
   io_msg.das_da_set_range.request     = _DAS_DA_SET_RANGE; 
   io_msg.das_da_set_range.fd          = fd;
   io_msg.das_da_set_range.chnl        = chnl;
   io_msg.das_da_set_range.range_min   = range_min;
   io_msg.das_da_set_range.range_max   = range_max;
   
   if(Sendfd(fd, &io_msg, &io_msg, sizeof(io_msg.das_da_set_range), 
               sizeof(io_msg.das_da_set_range_reply)) == -1)
      return(-1);
   if(io_msg.status != EOK)
   {
      errno = io_msg.status;
      return(-1);
   }
   return(0);
}

//*****************************************************************************
// das_da_get_trig - Send a message to get the current trigger source for D/A
//                   operations.
//*****************************************************************************
int das_da_get_trig(int fd, int chnl, int *trig_src)
{
                     // Fill in message structure.
   io_msg.das_da_get_trig.type    = _IO_DAS;
   io_msg.das_da_get_trig.request = _DAS_DA_GET_TRIG; 
   io_msg.das_da_get_trig.fd      = fd;
   io_msg.das_da_get_trig.chnl    = chnl;

   if(Sendfd(fd, &io_msg, &io_msg, sizeof(io_msg.das_da_get_trig), 
               sizeof(io_msg.das_da_set_trig_reply)) == -1)
      return(-1);
   if(io_msg.status != EOK)
   {
      errno = io_msg.status;
      return(-1);
   }
   *trig_src = io_msg.das_da_get_trig_reply.trig_src;
   return(0);
}

//*****************************************************************************
// das_da_set_trig - Send message to set a new trigger source for D/A 
//                   operations.
//*****************************************************************************
int das_da_set_trig(int fd, int chnl, int trig_src)
{
                     // Fill in message structure.
   io_msg.das_da_set_trig.type     = _IO_DAS;
   io_msg.das_da_set_trig.request  = _DAS_DA_SET_TRIG; 
   io_msg.das_da_set_trig.fd       = fd;
   io_msg.das_da_set_trig.trig_src = trig_src;
   io_msg.das_da_set_trig.chnl     = chnl;

   if(Sendfd(fd, &io_msg, &io_msg, sizeof(io_msg.das_da_set_trig), 
               sizeof(io_msg.das_da_set_trig_reply)) == -1)
      return(-1);
   if(io_msg.status != EOK)
   {
      errno = io_msg.status;
      return(-1);
   }
   return(0);
}

//*****************************************************************************
// das_da_set_sample -  Send message to set the sampling rate for D/A 
//                      operations.
//*****************************************************************************
int das_da_set_sample(int fd, int chnl, long ticks)
{
                     // Fill in message structure.
   io_msg.das_da_sam.type    = _IO_DAS;
   io_msg.das_da_sam.request = _DAS_DA_SET_SMP; 
   io_msg.das_da_sam.fd      = fd;
   io_msg.das_da_sam.ticks  = ticks;
   io_msg.das_da_sam.chnl    = chnl;

   if(Sendfd(fd, &io_msg, &io_msg, sizeof(io_msg.das_da_sam), 
               sizeof(io_msg.das_da_sam_reply)) == -1)
      return(-1);
   if(io_msg.status != EOK)
   {
      errno = io_msg.status;
      return(-1);
   }
   return(0);
}
