//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
// das_mem.c module contains routines to:
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
//    allocate DMA SAFE memory         - das_alloc()
//    deallocate far memory            - das_free_mem()
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

#include <i86.h>
#include <errno.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/uio.h>
#include <sys/types.h>
#include <sys/seginfo.h>
#include <sys/kernel.h>
#include <sys/mman.h>
#include <sys/shm.h>
#include <das.h>

// Function das_alloc() allocates nbytes of memory and returns a far pointer
// to it.
das_buf_t * das_alloc(int nbufs, int nbytes, unsigned flags)
{
   int fd, cnt, sel, n, i;
   unsigned nblks;
   caddr_t bufp;
   das_buf_t *cp, *dend, *dptr;
   struct _seginfo _seginfo;
      
   if(nbytes > 0xFFFF)
   {
      errno = EINVAL;
      return(NULL);
   }

   if((dptr = (das_buf_t *)calloc(1 + nbufs * sizeof(das_buf_t))) == NULL)
      return(NULL);

   if(!flags)                 // plain old memory
   {
      for(cnt = 0; cnt < nbufs; cnt++)
         if((dptr[cnt].bufp = (caddr_t)malloc(nbytes)) == NULL)
         {
            das_free(dptr);
            return(NULL);
         }
      return(dptr);
   }
   
   n = nbufs;
   dend = dptr+nbufs;
   for (cp = dptr; cp < dend; )
   {
      if ((nblks = n * nbytes) > 0xFFFF)
         nblks = 0xFFFF - (0xFFFF % nbytes);

      if ((sel = qnx_segment_alloc_flags(nblks, _PMF_DMA_SAFE)) == 0
            || sel == -1)
      {
         das_free(dptr);
         return(NULL);
      }

      qnx_segment_info(PROC_PID, getpid(), sel, &_seginfo);

#ifdef __386__
      if((fd = shm_open("Physical", O_RDWR, 0)) == -1)
      {
         das_free(dptr);
         return(NULL);
      }
      bufp = (caddr_t)mmap(0, nbytes, PROT_READ | PROT_WRITE, MAP_SHARED, fd, _seginfo.addr);
      if(bufp == NULL)
      {
         das_free(dptr);
         return(NULL);
      }
      close(fd);
#else
      bufp = MK_FP(sel, 0);
#endif
      
      for (i = 0; i < nblks/nbytes && cp < dend; ++i, ++cp, --n)
      {
         cp->sel                 = sel;                        // Segment 
         cp->bufp = cp->dma.bufp = i*nbytes+bufp;              // Offset
         cp->dma.phys_addr       = i*nbytes+_seginfo.addr;     // Physical
      }
   }
   return(dptr);
}

int das_free(das_buf_t *dptr)
{
   int cnt = 0, cur_sel;
   struct _seginfo _seginfo;
   
   if(dptr->dma.phys_addr == 0)              // free plain old memory
   {
      while(dptr[cnt].bufp)
         free(dptr[cnt++].bufp);
      free(dptr);
      return(EOK);
   }

   cur_sel = dptr[cnt].sel;
   while(dptr[cnt].bufp)
   {
#ifdef __386__
      if(cur_sel != dptr[cnt].sel || cnt == 0)
      {
         qnx_segment_info(PROC_PID, getpid(), cur_sel, &_seginfo);
         munmap(dptr[cnt].bufp, _seginfo.nbytes);
      }
#endif
      if(cur_sel != dptr[cnt].sel || cnt == 0)
         qnx_segment_free(dptr[cnt].sel);
      cur_sel = dptr[cnt].sel;
      cnt++;
   }
   free(dptr);
   return(EOK);
}
