// Advantech PCL-818 adapter
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
// das_open.c module contains routines to:
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
//    initialize the library/adapter   - das_open()
//    de-init the library/adapter      - das_close()
//    initialize digital I/O struct    - das_dio_init()
//    setup a default scan sequence    - das_dflt_seq()
//    initialize the cntr/tmr struct   - das_cntr_init()
//    initialize the A/D channels      - das_ad_init()
//    initialize the D/A channels      - das_da_init()
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

#include <sys/irqinfo.h>
#include <sys/proxy.h>
#include <signal.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <sys/fd.h>
#include <conio.h>
#include <unistd.h>
#include <i86.h>
#include <stdio.h>
#include <pcl_818.h>
#include <das.h>
#include <das_struct.h>
#include <das_extern.h>

int   init_flag = FALSE;            // library initialization status
int   int_id;                       // interrupt id
int _das_fd;                        // file descriptor returned from system
volatile struct docb docb;          // das operation control block

#pragma off (unreferenced)

// Function das_open() initializes the das library and adapter.  Paramaters
// passed in are checked for validity.  Proxies are created for event
// notification, a file descriptor is created, an interrupt handler is
// attached and structure initialization is performed.  An 'fd' is
// returned, and is checked for validity by all library routines.  The
// '*path' and 'oflag' paramaters are not used by the library routines but
// are used by the Resource Manager.
int das_open(const char *path, int oflag, const char *cfg_file)
{
   if(init_flag)                       // allow only one open 
   {
      errno = EBUSY;
      return(-1);
   }
   init_flag = TRUE;
   
   if(das_read_cfg_file(cfg_file) == -1)// read in the adapter configuration file
      return(-1);

   if(das_lib_init() == -1)            // setup interrupt handler, fd, proxies
      return(-1);

   das_dio_init();                     // initialize digital I/O structures
   das_ad_init();                      // initialize the A/D channels
   das_da_init();                      // initialize the D/A channels
   das_dflt_seq();                     // setup a default scan sequence
   das_cntr_init();                    // initialize the cntr/tmr structures
   das_catch_sig();                    // setup signal handler
   outp(docb.info.port+ADC_CTL, 0x00); // put the adapter in an inactive state
   return(_das_fd);                    // return the fd
}
#pragma on (unreferenced)

// Function das_dio_init() initializes the digital I/O structures
int das_dio_init()
{
   docb.din[0].bits  = 0;
   docb.din[0].addr  = P0_ADDR;
   docb.din[1].bits  = 0;
   docb.din[1].addr  = P1_ADDR;
   docb.dout[0].bits = 0;
   docb.dout[0].addr = P0_ADDR;
   docb.dout[1].bits = 0;
   docb.dout[1].addr = P1_ADDR;
   return(EOK);
}

// Function das_cntr_init() initializes the counter/timer structures
int das_cntr_init()
{
   int cnt;
   
   for(cnt = 0; cnt < MAX_CTR; cnt++)
   {
      docb.cntrs.ctr[cnt].md.ctl = 0;
      docb.cntrs.ctr[cnt].md.dir = 0;
      docb.cntrs.ctr[cnt].md.cmd = 0;
      docb.cntrs.ctr[cnt].md.md  = RATE_GEN; // only field used by 8254
      docb.cntrs.ctr[cnt].md.src = 0;
      docb.cntrs.ctr[cnt].md.edg = 0;
      docb.cntrs.ctr[cnt].md.gtl = 0;
      docb.cntrs.ctr[cnt].addr = PIT_CNT_0+cnt;
   }
   return(EOK);
}

// Function das_ad_init() initializes the A/D channel structures
int das_ad_init()
{
   int   cnt;

   for(cnt = 0; cnt < AD_CHNLS; cnt++)
   {
      docb.ad[cnt].stat       = COMPLETE;
      docb.ad[cnt].trig_src   = _PC_TRIG;
      docb.ad[cnt].addr       = ADC_DATA;
      docb.ad[cnt].rng        = _UN_X10;
      docb.ad[cnt].n_rent     = NULL;
      docb.ad[cnt].n_fent     = NULL;
      docb.ad[cnt].f_hd       = NULL;
      docb.ad[cnt].f_tl       = NULL;
      docb.ad[cnt].r_hd       = NULL;
      docb.ad[cnt].r_tl       = NULL;
      das_ad_set_sample(_das_fd, cnt, 1000);
   }
   return(EOK);
}

// Function das_da_init() initializes the D/A channel structures
int das_da_init()
{
   int cnt;

   for(cnt = 0; cnt < DA_CHNLS; cnt++)
   {
      docb.da[cnt].stat       = COMPLETE;
      docb.da[cnt].trig_src   = _PC_TRIG;
      docb.da[cnt].addr       = DAC0;
      docb.da[cnt].rng        = _UN_X10;
      docb.da[cnt].n_rent     = NULL;
      docb.da[cnt].n_fent     = NULL;
      docb.da[cnt].f_hd       = NULL;
      docb.da[cnt].f_tl       = NULL;
      docb.da[cnt].r_hd       = NULL;
      docb.da[cnt].r_tl       = NULL;
      das_da_set_sample(_das_fd, cnt, 1000);
   }
   docb.da[1].addr         = DAC1;
   return(EOK);
}

// Function das_dflt_seq() sets a default A/D scan sequence
int das_dflt_seq()
{
   int n_chnls;
   char seq[MAX_SNGL_CHNL+1] = { 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15 };

   n_chnls = ((docb.info.chnl_cnfg == DAS_SINGLE_ENDED) ? MAX_SNGL_CHNL : MAX_DIF_CHNL) + 1;
   das_ad_set_scan(_das_fd, &seq, n_chnls);
   return(EOK);
}

// Function das_close() shuts the adapter referenced by 'fd' down.  It
// terminates all A/D and D/A requests, releases it's 'fd', proxies, and
// IRQ.
int das_close(int fd)
{
   if(das_chk_init_fd(fd) != EOK)
      return(-1);
   das_ad_term(fd, 0);
   das_da_term(fd, 0);
   qnx_fd_detach(fd);                  // release the fd, proxies and handler
   qnx_hint_detach(int_id);            
   init_flag = FALSE;
   return(EOK);
}

// Function das_catch_sig() allows only SIGTERM, SIGKILL and SIGINT.  It
// ignores the rest.
void das_catch_sig()
{
   int i;
   for ( i = _SIGMIN; i <= _SIGMAX; ++i)
      signal( i, SIG_IGN);

   signal( SIGTERM, &das_terminate );
   signal( SIGKILL, &das_terminate );
   signal( SIGINT, &das_terminate );
}  

// Function das_terminate() handles SIGTERM, SIGKILL and SIGINT.
void das_terminate()
{
   das_close(_das_fd);
   exit(-1);
}

// Function das_get_info returns the current hardware configuration of the
// adapter.
int das_get_info(int fd, das_info *info)
{
   if(das_chk_init_fd(fd) != EOK)
      return(-1);
   memcpy(info, &docb.info, sizeof(das_info));
   return(EOK);
}

// Function das_read_cfg_file reads in the adapter specific configuration
// file.
int das_read_cfg_file(const char *cfg_file)
{
   FILE *fp;
   char buf[255] = CFG_FILE_PATH;

   if(cfg_file == NULL)
      strcat(buf, CFG_FILE);
   else
      strcat(buf, cfg_file);
      
   if((fp = fopen(buf, "r")) == NULL)
      return(-1);
      
   fgets(buf, 80, fp);
   docb.info.port = atoh(buf);
   fgets(buf, 80, fp);
   docb.info.irq = atoi(buf);
   fgets(buf, 80, fp);
   docb.info.dma_chnl = atoi(buf);
   fgets(buf, 80, fp);
   docb.info.range_ctrl = atoi(buf);
   fgets(buf, 80, fp);
   docb.info.chnl_cnfg = atoi(buf);
   fgets(buf, 80, fp);
   docb.info.pace_clk = atol(buf);
   fgets(buf, 80, fp);
   docb.info.ad_res = atoi(buf);
   fgets(buf, 80, fp);
   docb.info.da_res = atoi(buf);
   if(das_verify_cfg() == -1)
      return(-1);
   
   fclose(fp);
   return(EOK);
}

// Function das_verify_cfg verifies the configuration file.
das_verify_cfg()
{
                     // check for a valid DMA channel (1 or 3)
   if (docb.info.dma_chnl > MAX_DMA_CHNL || docb.info.dma_chnl < MIN_DMA_CHNL)
   {
      errno = EINVAL;
      return(-1);
   }
                                    // check for valid base addresses   
   if (docb.info.port > MAX_ADDR || docb.info.port < MIN_ADDR)
   {
      errno = EINVAL;
      return(-1);
   }
                     // check for a valid IRQ and attach an interrupt handler
   if ((docb.info.irq < MIN_IRQ) || (docb.info.irq > MAX_IRQ))
   {
      errno = EINVAL;
      return(-1);
   }
   return(EOK);
}

// Function das_lib_init attaches the interrupt handler, gets a file descriptor
// from the system, and creates a proxy for notification of AD and DA complete.
int das_lib_init()
{
      
   if ((int_id = qnx_hint_attach(docb.info.irq, &das_handler, FP_SEG(&docb))) == -1)
      return(-1);

                           // get an fd from the system
   if((_das_fd = qnx_fd_attach(0, 0, 0, getpid(), getpid(), 0, 0)) == -1)
      return(-1);

   return(EOK);
}

