// Advantech PCL-818 adapter
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
// das_tmr.c module contains routines to:
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
//    set a cntr to a specific mode    - das_ct_mode()
//    get the status of a cntr         - das_ct_status()
//    read the current count of a cntr - das_ct_read()
//    write the count to a cntr        - das_ct_write()           
//    stop the counting on a cntr      - das_ct_stop()
//    convert Hz to cntr values        - das_convert_rate()
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

#include <pcl_818.h>
#include <das_struct.h>
#include <das_extern.h>
#include <das_tmr.h>
#include <stdio.h>
#include <errno.h>
#include <conio.h>

// Function das_ct_mode() initializes a counter to a mode specified by
// 'md' in the cntr_mdr_t structure. 'n_ctr' is a value between 1 and 3 which 
// directly relates to the 3 timers available on the 8254 timer/counter 
// chip.  'clk_src' is used to specify the clock source for counter 1, it can 
// be either INTERNAL (100kHz) or EXTERNAL.
// 
// NOTE : PCL818 - Counter 0 is the only counter with a configurable clock
//        source.
//
#pragma off (check_stack)
#pragma off (unreferenced)
int das_ct_mode(int fd, int n_ctr, cntr_mdr_t *cptr, int clk_src)
{
   if (n_ctr > MAX_CTR)
   {
      errno = EINVAL;
      return(-1);
   }
   outp(docb.info.port+PIT_CTL, RW_LM | (cptr->md << 1) | ((n_ctr-1) << 6));

   if (clk_src != NULL)
   {
      if (n_ctr != 1)
      {
         errno = EINVAL;
         return(-1);
      }
      clk_src &= 2;
      docb.cntrs.pit_stat &= clk_src;
      outp(docb.info.port+PIT_ENABLE, docb.cntrs.pit_stat);
   }
   return(EOK);
}   
 
// Function das_ct_status() returns the status of the counter specified
// by 'n_ctr' in 'status'.  Contents of status are as follows :
//                         d7: OUT  - Counter output current state.
//                         d6: NC   - Null count.
//                         d5: RW1  -\ Read/Write operation preformed 
//                         d4: RW0  -/ on counter.
//                         d3: M2   -\   
//                         d2: M1   -> Counter mode.
//                         d1: M0   -/
//                         d0: BCD  - Counting Mode.
//
int das_ct_status(int fd, int n_ctr, int *status)
{
   if (n_ctr > MAX_CTR)  
   {
      errno = EINVAL;
      return(-1);
   }
   
   outp(docb.info.port+PIT_CTL, RB_MODE); // RB_MODE = 11000000
   outp(docb.info.port+PIT_CTL, G_STAT| (CTR << (n_ctr-1))); // CTR = 1
   
   *status =inp(docb.info.port+docb.cntrs.ctr[n_ctr-1].addr);
   return(EOK);
}

// Function das_ct_read() returns the current count of the counter
// specified by 'n_ctr'. 
int das_ct_read(int fd, int n_ctr, int *cnt)
{
   if (n_ctr > MAX_CTR)
   {
      errno = EINVAL;
      return(-1);
   }
   
   outp(docb.info.port+PIT_CTL, RB_MODE); // RB_MODE = 1100000
   outp(docb.info.port+PIT_CTL, G_CNT | (CTR << (n_ctr-1))); // CTR = 1
   
   *cnt = inp(docb.info.port+docb.cntrs.ctr[n_ctr-1].addr);
   *cnt |= inp(docb.info.port+docb.cntrs.ctr[n_ctr-1].addr) << 8;
   return(EOK);
}

// Function das_ct_write() writes the count specified by 'value' to the
// counter specified by 'n_ctr'.
int das_ct_write(int fd, int n_ctr, cntr_t *cptr)
{
   if (n_ctr > MAX_CTR)
   {
      errno = EINVAL;
      return(-1);
   }
   
   outp(docb.info.port+docb.cntrs.ctr[n_ctr-1].addr, cptr->ld);
   outp(docb.info.port+docb.cntrs.ctr[n_ctr-1].addr, cptr->ld >> 8);
   return(EOK);
}

// Function das_ct_stop() stops counting operation on counter specified
// by 'n_ctr'.
int das_ct_stop(int fd, int n_ctr)
{
   if(n_ctr > MAX_CTR)
   {
      errno = EINVAL;
      return(-1);
   }
   outp(docb.info.port+PIT_CTL, ((n_ctr-1) << 6) | RW_LM); 
   return(EOK);
}
#pragma on (unreferenced)
#pragma on (check_stack)
